# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import time
import random

from ..log import logger
from .profile import Profile
from .profile import CallSite
from .schedule import schedule, delay


class SamplerConfig(object):
    def __init__(self):
        self.log_prefix = None
        self.max_profile_duration = None
        self.max_span_duration = None
        self.span_interval = None
        self.report_interval = None


class SamplerScheduler:
    def __init__(self, profiler, sampler, config):
        self.profiler = profiler
        self.sampler = sampler
        self.config = config
        self.started = False
        self.span_timer = None
        self.span_timeout = None
        self.random_timer = None
        self.report_timer = None
        self.profile_start_ts = None
        self.profile_duration = None
        self.span_active = False
        self.span_start_ts = None
        self.span_count = 0

    def setup(self):
        self.sampler.setup()

    def start(self):
        if not self.sampler.ready:
            return

        if self.started:
            return
        self.started = True

        self.reset()

        def random_delay():
            timeout = random.randint(0, round(self.config.span_interval - self.config.max_span_duration))
            self.random_timer = delay(timeout, self.start_profiling)

        if not self.profiler.get_option('disable_timers'):
            self.span_timer = schedule(0, self.config.span_interval, random_delay)
            self.report_timer = schedule(self.config.report_interval, self.config.report_interval, self.report)

    def stop(self):
        if not self.started:
            return

        self.started = False

        if self.span_timer:
            self.span_timer.cancel()
            self.span_timer = None

        if self.random_timer:
            self.random_timer.cancel()
            self.random_timer = None

        if self.report_timer:
            self.report_timer.cancel()
            self.report_timer = None

        self.stop_profiling()

    def destroy(self):
        self.sampler.destroy()

    def reset(self):
        self.sampler.reset()
        self.profile_start_ts = time.time()
        self.profile_duration = 0
        self.span_count = 0

    def start_profiling(self):
        if not self.started:
            return False

        if self.profile_duration > self.config.max_profile_duration:
            logger.debug(self.config.log_prefix + ': max profiling duration reached.')
            return False

        if self.span_count > self.config.max_span_count:
            logger.debug(self.config.log_prefix + ': max recording count reached.')
            return False

        if self.profiler.sampler_active:
            logger.debug(self.config.log_prefix + ': sampler lock exists.')
            return False
        self.profiler.sampler_active = True
        logger.debug(self.config.log_prefix + ': started.')

        try:
            self.sampler.start_sampler()
        except Exception:
            self.profiler.sampler_active = False
            logger.error('Error starting profiling', exc_info=True)
            return False

        self.span_timeout = delay(self.config.max_span_duration, self.stop_profiling)
        
        self.span_active = True
        self.span_start_ts = time.time()
        self.span_count += 1

        return True

    def stop_profiling(self):
        if not self.span_active:
            return
        self.span_active = False

        try:
            self.profile_duration = self.profile_duration + time.time() - self.span_start_ts
            self.sampler.stop_sampler()
        except Exception:
            logger.error('Error stopping profiling', exc_info=True)

        self.profiler.sampler_active = False

        if self.span_timeout:
            self.span_timeout.cancel()

        logger.debug(self.config.log_prefix + ': stopped.')

    def report(self):
        if not self.started:
            return

        if self.profile_duration == 0:
            return

        if self.profile_start_ts > time.time() - self.config.report_interval:
            return
        elif self.profile_start_ts < time.time() - 2 * self.config.report_interval:
            self.reset()
            return

        profile = self.sampler.build_profile(
            to_millis(self.profile_duration), 
            to_millis(time.time() - self.profile_start_ts))

        if self.profiler.agent.can_send():
            if self.profiler.agent.announce_data.pid:
                profile.process_id = str(self.profiler.agent.announce_data.pid)

            self.profiler.agent.collector.profile_queue.put(profile.to_dict())

            logger.debug(self.config.log_prefix + ': reporting profile:')
        else:
            logger.debug(self.config.log_prefix + ': not reporting profile, agent not ready')

        self.reset()


def to_millis(t):
    return int(round(t * 1000))



# coding: utf-8
# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021


from __future__ import absolute_import

import wrapt
import opentracing
import types

from ..log import logger
from ..singletons import tracer
from ..util.traceutils import get_active_tracer

try:
    import pika


    def _extract_broker_tags(span, conn):
        span.set_tag("address", "%s:%d" % (conn.params.host, conn.params.port))


    def _extract_publisher_tags(span, conn, exchange, routing_key):
        _extract_broker_tags(span, conn)

        span.set_tag("sort", "publish")
        span.set_tag("key", routing_key)
        span.set_tag("exchange", exchange)


    def _extract_consumer_tags(span, conn, queue):
        _extract_broker_tags(span, conn)

        span.set_tag("address", "%s:%d" % (conn.params.host, conn.params.port))
        span.set_tag("sort", "consume")
        span.set_tag("queue", queue)


    @wrapt.patch_function_wrapper('pika.channel', 'Channel.basic_publish')
    def basic_publish_with_instana(wrapped, instance, args, kwargs):
        def _bind_args(exchange, routing_key, body, properties=None, *args, **kwargs):
            return (exchange, routing_key, body, properties, args, kwargs)

        active_tracer = get_active_tracer()

        if active_tracer is None:
            return wrapped(*args, **kwargs)

        (exchange, routing_key, body, properties, args, kwargs) = (_bind_args(*args, **kwargs))

        with tracer.start_active_span("rabbitmq", child_of=active_tracer.active_span) as scope:
            try:
                _extract_publisher_tags(scope.span,
                                        conn=instance.connection,
                                        routing_key=routing_key,
                                        exchange=exchange)
            except:
                logger.debug("publish_with_instana: ", exc_info=True)

            # context propagation
            properties = properties or pika.BasicProperties()
            properties.headers = properties.headers or {}

            tracer.inject(scope.span.context, opentracing.Format.HTTP_HEADERS, properties.headers,
                          disable_w3c_trace_context=True)
            args = (exchange, routing_key, body, properties) + args

            try:
                rv = wrapped(*args, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return rv


    def basic_get_with_instana(wrapped, instance, args, kwargs):
        def _bind_args(queue, callback, *args, **kwargs):
            return (queue, callback, args, kwargs)

        queue, callback, args, kwargs = _bind_args(*args, **kwargs)

        def _cb_wrapper(channel, method, properties, body):
            parent_span = tracer.extract(opentracing.Format.HTTP_HEADERS, properties.headers,
                                         disable_w3c_trace_context=True)

            with tracer.start_active_span("rabbitmq", child_of=parent_span) as scope:
                try:
                    _extract_consumer_tags(scope.span,
                                           conn=instance.connection,
                                           queue=queue)
                except:
                    logger.debug("basic_get_with_instana: ", exc_info=True)

                try:
                    callback(channel, method, properties, body)
                except Exception as e:
                    scope.span.log_exception(e)
                    raise

        args = (queue, _cb_wrapper) + args
        return wrapped(*args, **kwargs)


    @wrapt.patch_function_wrapper('pika.adapters.blocking_connection', 'BlockingChannel.basic_consume')
    def basic_consume_with_instana(wrapped, instance, args, kwargs):
        def _bind_args(queue, on_consume_callback, *args, **kwargs):
            return (queue, on_consume_callback, args, kwargs)

        queue, on_consume_callback, args, kwargs = _bind_args(*args, **kwargs)

        def _cb_wrapper(channel, method, properties, body):
            parent_span = tracer.extract(opentracing.Format.HTTP_HEADERS, properties.headers,
                                         disable_w3c_trace_context=True)

            with tracer.start_active_span("rabbitmq", child_of=parent_span) as scope:
                try:
                    _extract_consumer_tags(scope.span,
                                           conn=instance.connection,
                                           queue=queue)
                except:
                    logger.debug("basic_consume_with_instana: ", exc_info=True)

                try:
                    on_consume_callback(channel, method, properties, body)
                except Exception as e:
                    scope.span.log_exception(e)
                    raise

        args = (queue, _cb_wrapper) + args
        return wrapped(*args, **kwargs)


    @wrapt.patch_function_wrapper('pika.adapters.blocking_connection', 'BlockingChannel.consume')
    def consume_with_instana(wrapped, instance, args, kwargs):
        def _bind_args(queue, *args, **kwargs):
            return (queue, args, kwargs)

        (queue, args, kwargs) = (_bind_args(*args, **kwargs))

        def _consume(gen):
            for yilded in gen:
                # Bypass the delivery created due to inactivity timeout
                if yilded is None or not any(yilded):
                    yield yilded
                    continue

                (method_frame, properties, body) = yilded

                parent_span = tracer.extract(opentracing.Format.HTTP_HEADERS, properties.headers,
                                             disable_w3c_trace_context=True)
                with tracer.start_active_span("rabbitmq", child_of=parent_span) as scope:
                    try:
                        _extract_consumer_tags(scope.span,
                                               conn=instance.connection._impl,
                                               queue=queue)
                    except:
                        logger.debug("consume_with_instana: ", exc_info=True)

                    try:
                        yield yilded
                    except Exception as e:
                        scope.span.log_exception(e)
                        raise

        args = (queue,) + args
        res = wrapped(*args, **kwargs)

        if isinstance(res, types.GeneratorType):
            return _consume(res)
        else:
            return res


    @wrapt.patch_function_wrapper('pika.adapters.blocking_connection', 'BlockingChannel.__init__')
    def _BlockingChannel___init__(wrapped, instance, args, kwargs):
        ret = wrapped(*args, **kwargs)
        impl = getattr(instance, '_impl', None)

        if impl and hasattr(impl.basic_consume, '__wrapped__'):
            impl.basic_consume = impl.basic_consume.__wrapped__

        return ret


    wrapt.wrap_function_wrapper('pika.channel', 'Channel.basic_get', basic_get_with_instana)
    wrapt.wrap_function_wrapper('pika.channel', 'Channel.basic_consume', basic_get_with_instana)

    logger.debug("Instrumenting pika")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

from __future__ import absolute_import

import opentracing
import wrapt

from ...log import logger
from ...singletons import agent, async_tracer
from ...util.secrets import strip_secrets_from_query

try:
    import aiohttp
    import asyncio


    async def stan_request_start(session, trace_config_ctx, params):
        try:
            parent_span = async_tracer.active_span

            # If we're not tracing, just return
            if parent_span is None:
                trace_config_ctx.scope = None
                return

            scope = async_tracer.start_active_span("aiohttp-client", child_of=parent_span)
            trace_config_ctx.scope = scope

            async_tracer.inject(scope.span.context, opentracing.Format.HTTP_HEADERS, params.headers)

            parts = str(params.url).split('?')
            if len(parts) > 1:
                cleaned_qp = strip_secrets_from_query(parts[1], agent.options.secrets_matcher,
                                                      agent.options.secrets_list)
                scope.span.set_tag("http.params", cleaned_qp)
            scope.span.set_tag("http.url", parts[0])
            scope.span.set_tag('http.method', params.method)
        except Exception:
            logger.debug("stan_request_start", exc_info=True)


    async def stan_request_end(session, trace_config_ctx, params):
        try:
            scope = trace_config_ctx.scope
            if scope is not None:
                scope.span.set_tag('http.status_code', params.response.status)

                if agent.options.extra_http_headers is not None:
                    for custom_header in agent.options.extra_http_headers:
                        if custom_header in params.response.headers:
                            scope.span.set_tag("http.header.%s" % custom_header, params.response.headers[custom_header])

                if 500 <= params.response.status <= 599:
                    scope.span.mark_as_errored({"http.error": params.response.reason})

                scope.close()
        except Exception:
            logger.debug("stan_request_end", exc_info=True)


    async def stan_request_exception(session, trace_config_ctx, params):
        try:
            scope = trace_config_ctx.scope
            if scope is not None:
                scope.span.log_exception(params.exception)
                scope.span.set_tag("http.error", str(params.exception))
                scope.close()
        except Exception:
            logger.debug("stan_request_exception", exc_info=True)


    @wrapt.patch_function_wrapper('aiohttp.client', 'ClientSession.__init__')
    def init_with_instana(wrapped, instance, argv, kwargs):
        instana_trace_config = aiohttp.TraceConfig()
        instana_trace_config.on_request_start.append(stan_request_start)
        instana_trace_config.on_request_end.append(stan_request_end)
        instana_trace_config.on_request_exception.append(stan_request_exception)
        if 'trace_configs' in kwargs:
            kwargs['trace_configs'].append(instana_trace_config)
        else:
            kwargs['trace_configs'] = [instana_trace_config]

        return wrapped(*argv, **kwargs)


    logger.debug("Instrumenting aiohttp client")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

""" Module to handle the collection of Docker metrics in AWS Fargate """
from __future__ import division
from ....log import logger
from ..base import BaseHelper
from ....util import DictionaryOfStan


class DockerHelper(BaseHelper):
    """ This class acts as a helper to collect Docker snapshot and metric information """
    def __init__(self, collector):
        super(DockerHelper, self).__init__(collector)

        # The metrics from the previous report cycle
        self.previous = DictionaryOfStan()

        # For metrics that are accumalative, store their previous values here
        # Indexed by docker_id:  self.previous_blkio[docker_id][metric]
        self.previous_blkio = DictionaryOfStan()

    def collect_metrics(self, **kwargs):
        """
        Collect and return docker metrics (and optionally snapshot data) for this task
        @return: list - with one or more plugin entities
        """
        plugins = []
        try:
            if self.collector.task_metadata is not None:
                containers = self.collector.task_metadata.get("Containers", [])
                for container in containers:
                    plugin_data = dict()
                    plugin_data["name"] = "com.instana.plugin.docker"
                    docker_id = container.get("DockerId")

                    name = container.get("Name", "")
                    labels = container.get("Labels", {})
                    task_arn = labels.get("com.amazonaws.ecs.task-arn", "")

                    plugin_data["entityId"] = "%s::%s" % (task_arn, name)
                    plugin_data["data"] = DictionaryOfStan()
                    plugin_data["data"]["Id"] = container.get("DockerId", None)

                    with_snapshot = kwargs.get("with_snapshot", False)
                    # Metrics
                    self._collect_container_metrics(plugin_data, docker_id, with_snapshot)

                    # Snapshot
                    if with_snapshot:
                        self._collect_container_snapshot(plugin_data, container)

                    plugins.append(plugin_data)
                    #logger.debug(to_pretty_json(plugin_data))
        except Exception:
            logger.debug("DockerHelper.collect_metrics: ", exc_info=True)
        return plugins

    def _collect_container_snapshot(self, plugin_data, container):
        try:
            # Snapshot Data
            plugin_data["data"]["Created"] = container.get("CreatedAt", None)
            plugin_data["data"]["Started"] = container.get("StartedAt", None)
            plugin_data["data"]["Image"] = container.get("Image", None)
            plugin_data["data"]["Labels"] = container.get("Labels", None)
            plugin_data["data"]["Ports"] = container.get("Ports", None)

            networks = container.get("Networks", [])
            if len(networks) >= 1:
                plugin_data["data"]["NetworkMode"] = networks[0].get("NetworkMode", None)
        except Exception:
            logger.debug("_collect_container_snapshot: ", exc_info=True)

    def _collect_container_metrics(self, plugin_data, docker_id, with_snapshot):
        container = self.collector.task_stats_metadata.get(docker_id, None)
        if container is not None:
            self._collect_network_metrics(container, plugin_data, docker_id, with_snapshot)
            self._collect_cpu_metrics(container, plugin_data, docker_id, with_snapshot)
            self._collect_memory_metrics(container, plugin_data, docker_id, with_snapshot)
            self._collect_blkio_metrics(container, plugin_data, docker_id, with_snapshot)

    def _collect_network_metrics(self, container, plugin_data, docker_id, with_snapshot):
        try:
            networks = container.get("networks", None)
            tx_bytes_total = tx_dropped_total = tx_errors_total = tx_packets_total = 0
            rx_bytes_total = rx_dropped_total = rx_errors_total = rx_packets_total = 0

            if networks is not None:
                for key in networks.keys():
                    if "eth" in key:
                        tx_bytes_total += networks[key].get("tx_bytes", 0)
                        tx_dropped_total += networks[key].get("tx_dropped", 0)
                        tx_errors_total += networks[key].get("tx_errors", 0)
                        tx_packets_total += networks[key].get("tx_packets", 0)

                        rx_bytes_total += networks[key].get("rx_bytes", 0)
                        rx_dropped_total += networks[key].get("rx_dropped", 0)
                        rx_errors_total += networks[key].get("rx_errors", 0)
                        rx_packets_total += networks[key].get("rx_packets", 0)

                self.apply_delta(tx_bytes_total, self.previous[docker_id]["network"]["tx"],
                                 plugin_data["data"]["tx"], "bytes", with_snapshot)
                self.apply_delta(tx_dropped_total, self.previous[docker_id]["network"]["tx"],
                                 plugin_data["data"]["tx"], "dropped", with_snapshot)
                self.apply_delta(tx_errors_total, self.previous[docker_id]["network"]["tx"],
                                 plugin_data["data"]["tx"], "errors", with_snapshot)
                self.apply_delta(tx_packets_total, self.previous[docker_id]["network"]["tx"],
                                 plugin_data["data"]["tx"], "packets", with_snapshot)

                self.apply_delta(rx_bytes_total, self.previous[docker_id]["network"]["rx"],
                                 plugin_data["data"]["rx"], "bytes", with_snapshot)
                self.apply_delta(rx_dropped_total, self.previous[docker_id]["network"]["rx"],
                                 plugin_data["data"]["rx"], "dropped", with_snapshot)
                self.apply_delta(rx_errors_total, self.previous[docker_id]["network"]["rx"],
                                 plugin_data["data"]["rx"], "errors", with_snapshot)
                self.apply_delta(rx_packets_total, self.previous[docker_id]["network"]["rx"],
                                 plugin_data["data"]["rx"], "packets", with_snapshot)
        except Exception:
            logger.debug("_collect_network_metrics: ", exc_info=True)

    def _collect_cpu_metrics(self, container, plugin_data, docker_id, with_snapshot):
        try:
            cpu_stats = container.get("cpu_stats", {})
            cpu_usage = cpu_stats.get("cpu_usage", None)
            throttling_data = cpu_stats.get("throttling_data", None)

            if cpu_usage is not None:
                online_cpus = cpu_stats.get("online_cpus", 1)
                system_cpu_usage = cpu_stats.get("system_cpu_usage", 0)

                metric_value = (cpu_usage["total_usage"] / system_cpu_usage) * online_cpus
                self.apply_delta(round(metric_value, 6),
                                 self.previous[docker_id]["cpu"],
                                 plugin_data["data"]["cpu"], "total_usage", with_snapshot)

                metric_value = (cpu_usage["usage_in_usermode"] / system_cpu_usage) * online_cpus
                self.apply_delta(round(metric_value, 6),
                                 self.previous[docker_id]["cpu"],
                                 plugin_data["data"]["cpu"], "user_usage", with_snapshot)

                metric_value = (cpu_usage["usage_in_kernelmode"] / system_cpu_usage) * online_cpus
                self.apply_delta(round(metric_value, 6),
                                 self.previous[docker_id]["cpu"],
                                 plugin_data["data"]["cpu"], "system_usage", with_snapshot)

            if throttling_data is not None:
                self.apply_delta(throttling_data,
                                 self.previous[docker_id]["cpu"],
                                 plugin_data["data"]["cpu"], ("periods", "throttling_count"), with_snapshot)
                self.apply_delta(throttling_data,
                                 self.previous[docker_id]["cpu"],
                                 plugin_data["data"]["cpu"], ("throttled_time", "throttling_time"), with_snapshot)
        except Exception:
            logger.debug("_collect_cpu_metrics: ", exc_info=True)

    def _collect_memory_metrics(self, container, plugin_data, docker_id, with_snapshot):
        try:
            memory = container.get("memory_stats", {})
            memory_stats = memory.get("stats", None)

            self.apply_delta(memory, self.previous[docker_id]["memory"],
                             plugin_data["data"]["memory"], "usage", with_snapshot)
            self.apply_delta(memory, self.previous[docker_id]["memory"],
                             plugin_data["data"]["memory"], "max_usage", with_snapshot)
            self.apply_delta(memory, self.previous[docker_id]["memory"],
                             plugin_data["data"]["memory"], "limit", with_snapshot)

            if memory_stats is not None:
                self.apply_delta(memory_stats, self.previous[docker_id]["memory"],
                                 plugin_data["data"]["memory"], "active_anon", with_snapshot)
                self.apply_delta(memory_stats, self.previous[docker_id]["memory"],
                                 plugin_data["data"]["memory"], "active_file", with_snapshot)
                self.apply_delta(memory_stats, self.previous[docker_id]["memory"],
                                 plugin_data["data"]["memory"], "inactive_anon", with_snapshot)
                self.apply_delta(memory_stats, self.previous[docker_id]["memory"],
                                 plugin_data["data"]["memory"], "inactive_file", with_snapshot)
                self.apply_delta(memory_stats, self.previous[docker_id]["memory"],
                                 plugin_data["data"]["memory"], "total_cache", with_snapshot)
                self.apply_delta(memory_stats, self.previous[docker_id]["memory"],
                                 plugin_data["data"]["memory"], "total_rss", with_snapshot)
        except Exception:
            logger.debug("_collect_memory_metrics: ", exc_info=True)

    def _collect_blkio_metrics(self, container, plugin_data, docker_id, with_snapshot):
        try:
            blkio_stats = container.get("blkio_stats", None)
            if blkio_stats is not None:
                service_bytes = blkio_stats.get("io_service_bytes_recursive", None)
                if service_bytes is not None:
                    for entry in service_bytes:
                        if entry["op"] == "Read":
                            previous_value = self.previous_blkio[docker_id].get("blk_read", 0)
                            value_diff = entry["value"] - previous_value
                            self.apply_delta(value_diff, self.previous[docker_id]["blkio"],
                                             plugin_data["data"]["blkio"], "blk_read", with_snapshot)
                            self.previous_blkio[docker_id]["blk_read"] = entry["value"]
                        elif entry["op"] == "Write":
                            previous_value = self.previous_blkio[docker_id].get("blk_write", 0)
                            value_diff = entry["value"] - previous_value
                            self.apply_delta(value_diff, self.previous[docker_id]["blkio"],
                                             plugin_data["data"]["blkio"], "blk_write", with_snapshot)
                            self.previous_blkio[docker_id]["blk_write"] = entry["value"]
        except Exception:
            logger.debug("_collect_blkio_metrics: ", exc_info=True)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import os
import sys
import json
import time
import wrapt
import logging
import unittest

from instana.tracer import InstanaTracer
from instana.agent.aws_lambda import AWSLambdaAgent
from instana.options import AWSLambdaOptions
from instana.recorder import StanRecorder
from instana import lambda_handler
from instana import get_lambda_handler_or_default
from instana.instrumentation.aws.lambda_inst import lambda_handler_with_instana
from instana.instrumentation.aws.triggers import read_http_query_params
from instana.singletons import get_agent, set_agent, get_tracer, set_tracer
from instana.util.aws import normalize_aws_lambda_arn


# Mock Context object
class MockContext(dict):
    def __init__(self, **kwargs):
        super(MockContext, self).__init__(**kwargs)
        self.invoked_function_arn = "arn:aws:lambda:us-east-2:12345:function:TestPython:1"
        self.function_name = "TestPython"
        self.function_version = "1"


# This is the target handler that will be instrumented for these tests
def my_lambda_handler(event, context):
    # print("target_handler called")
    return {
        'statusCode': 200,
        'headers': {'Content-Type': 'application/json'},
        'body': json.dumps({'site': 'pwpush.com', 'response': 204})
    }

# We only want to monkey patch the test handler once so do it here
os.environ["LAMBDA_HANDLER"] = "tests.platforms.test_lambda.my_lambda_handler"
module_name, function_name = get_lambda_handler_or_default()
wrapt.wrap_function_wrapper(module_name, function_name, lambda_handler_with_instana)


class TestLambda(unittest.TestCase):
    def __init__(self, methodName='runTest'):
        super(TestLambda, self).__init__(methodName)
        self.agent = None
        self.span_recorder = None
        self.tracer = None
        self.pwd = os.path.dirname(os.path.realpath(__file__))

        self.original_agent = get_agent()
        self.original_tracer = get_tracer()

    def setUp(self):
        os.environ["AWS_EXECUTION_ENV"] = "AWS_Lambda_python_3.8"
        os.environ["LAMBDA_HANDLER"] = "tests.platforms.test_lambda.my_lambda_handler"
        os.environ["INSTANA_ENDPOINT_URL"] = "https://localhost/notreal"
        os.environ["INSTANA_AGENT_KEY"] = "Fake_Key"
        self.context = MockContext()

    def tearDown(self):
        """ Reset all environment variables of consequence """
        if "AWS_EXECUTION_ENV" in os.environ:
            os.environ.pop("AWS_EXECUTION_ENV")
        if "LAMBDA_HANDLER" in os.environ:
            os.environ.pop("LAMBDA_HANDLER")
        if "INSTANA_EXTRA_HTTP_HEADERS" in os.environ:
            os.environ.pop("INSTANA_EXTRA_HTTP_HEADERS")
        if "INSTANA_ENDPOINT_URL" in os.environ:
            os.environ.pop("INSTANA_ENDPOINT_URL")
        if "INSTANA_ENDPOINT_PROXY" in os.environ:
            os.environ.pop("INSTANA_ENDPOINT_PROXY")
        if "INSTANA_AGENT_KEY" in os.environ:
            os.environ.pop("INSTANA_AGENT_KEY")
        if "INSTANA_SERVICE_NAME" in os.environ:
            os.environ.pop("INSTANA_SERVICE_NAME")
        if "INSTANA_DEBUG" in os.environ:
            os.environ.pop("INSTANA_DEBUG")
        if "INSTANA_LOG_LEVEL" in os.environ:
            os.environ.pop("INSTANA_LOG_LEVEL")

        set_agent(self.original_agent)
        set_tracer(self.original_tracer)

    def create_agent_and_setup_tracer(self):
        self.agent = AWSLambdaAgent()
        self.span_recorder = StanRecorder(self.agent)
        self.tracer = InstanaTracer(recorder=self.span_recorder)
        set_agent(self.agent)
        set_tracer(self.tracer)

    def test_invalid_options(self):
        # None of the required env vars are available...
        if "LAMBDA_HANDLER" in os.environ:
            os.environ.pop("LAMBDA_HANDLER")
        if "INSTANA_EXTRA_HTTP_HEADERS" in os.environ:
            os.environ.pop("INSTANA_EXTRA_HTTP_HEADERS")
        if "INSTANA_ENDPOINT_URL" in os.environ:
            os.environ.pop("INSTANA_ENDPOINT_URL")
        if "INSTANA_AGENT_KEY" in os.environ:
            os.environ.pop("INSTANA_AGENT_KEY")

        agent = AWSLambdaAgent()
        self.assertFalse(agent._can_send)
        self.assertIsNone(agent.collector)

    def test_secrets(self):
        self.create_agent_and_setup_tracer()
        self.assertTrue(hasattr(self.agent.options, 'secrets_matcher'))
        self.assertEqual(self.agent.options.secrets_matcher, 'contains-ignore-case')
        self.assertTrue(hasattr(self.agent.options, 'secrets_list'))
        self.assertEqual(self.agent.options.secrets_list, ['key', 'pass', 'secret'])

    def test_has_extra_http_headers(self):
        self.create_agent_and_setup_tracer()
        self.assertTrue(hasattr(self.agent, 'options'))
        self.assertTrue(hasattr(self.agent.options, 'extra_http_headers'))

    def test_has_options(self):
        self.create_agent_and_setup_tracer()
        self.assertTrue(hasattr(self.agent, 'options'))
        self.assertTrue(isinstance(self.agent.options, AWSLambdaOptions))
        assert(self.agent.options.endpoint_proxy == { })

    def test_get_handler(self):
        os.environ["LAMBDA_HANDLER"] = "tests.lambda_handler"
        handler_module, handler_function = get_lambda_handler_or_default()

        self.assertEqual("tests", handler_module)
        self.assertEqual("lambda_handler", handler_function)

    def test_get_handler_with_multi_subpackages(self):
        os.environ["LAMBDA_HANDLER"] = "tests.one.two.three.lambda_handler"
        handler_module, handler_function = get_lambda_handler_or_default()

        self.assertEqual("tests.one.two.three", handler_module)
        self.assertEqual("lambda_handler", handler_function)

    def test_get_handler_with_space_in_it(self):
        os.environ["LAMBDA_HANDLER"] = " tests.another_module.lambda_handler"
        handler_module, handler_function = get_lambda_handler_or_default()

        self.assertEqual("tests.another_module", handler_module)
        self.assertEqual("lambda_handler", handler_function)

        os.environ["LAMBDA_HANDLER"] = "tests.another_module.lambda_handler    "
        handler_module, handler_function = get_lambda_handler_or_default()

        self.assertEqual("tests.another_module", handler_module)
        self.assertEqual("lambda_handler", handler_function)

    def test_agent_extra_http_headers(self):
        os.environ['INSTANA_EXTRA_HTTP_HEADERS'] = "X-Test-Header;X-Another-Header;X-And-Another-Header"
        self.create_agent_and_setup_tracer()
        self.assertIsNotNone(self.agent.options.extra_http_headers)
        should_headers = ['x-test-header', 'x-another-header', 'x-and-another-header']
        self.assertEqual(should_headers, self.agent.options.extra_http_headers)

    def test_custom_proxy(self):
        os.environ["INSTANA_ENDPOINT_PROXY"] = "http://myproxy.123"
        self.create_agent_and_setup_tracer()
        assert(self.agent.options.endpoint_proxy == { 'https': "http://myproxy.123" })

    def test_custom_service_name(self):
        os.environ['INSTANA_SERVICE_NAME'] = "Legion"
        with open(self.pwd + '/../data/lambda/api_gateway_event.json', 'r') as json_file:
            event = json.load(json_file)

        self.create_agent_and_setup_tracer()

        # Call the Instana Lambda Handler as we do in the real world.  It will initiate tracing and then
        # figure out the original (the users') Lambda Handler and execute it.
        # The original Lambda handler is set in os.environ["LAMBDA_HANDLER"]
        result = lambda_handler(event, self.context)
        os.environ.pop('INSTANA_SERVICE_NAME')

        assert isinstance(result, dict)
        assert 'headers' in result
        assert 'Server-Timing' in result['headers']

        time.sleep(1)
        payload = self.agent.collector.prepare_payload()

        self.assertTrue("metrics" in payload)
        self.assertTrue("spans" in payload)
        self.assertEqual(2, len(payload.keys()))

        self.assertTrue(isinstance(payload['metrics']['plugins'], list))
        self.assertTrue(len(payload['metrics']['plugins']) == 1)
        plugin_data = payload['metrics']['plugins'][0]

        self.assertEqual('com.instana.plugin.aws.lambda', plugin_data['name'])
        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', plugin_data['entityId'])

        self.assertEqual(1, len(payload['spans']))

        span = payload['spans'][0]
        self.assertEqual('aws.lambda.entry', span.n)
        self.assertEqual('d5cb361b256413a9', span.t)
        self.assertIsNotNone(span.s)
        self.assertEqual('0901d8ae4fbf1529', span.p)
        self.assertIsNotNone(span.ts)
        self.assertIsNotNone(span.d)

        server_timing_value = "intid;desc=%s" % span.t
        assert result['headers']['Server-Timing'] == server_timing_value

        self.assertEqual({'hl': True, 'cp': 'aws', 'e': 'arn:aws:lambda:us-east-2:12345:function:TestPython:1'},
                         span.f)

        self.assertTrue(span.sy)

        self.assertIsNone(span.ec)
        self.assertIsNone(span.data['lambda']['error'])

        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', span.data['lambda']['arn'])
        self.assertEqual(None, span.data['lambda']['alias'])
        self.assertEqual('python', span.data['lambda']['runtime'])
        self.assertEqual('TestPython', span.data['lambda']['functionName'])
        self.assertEqual('1', span.data['lambda']['functionVersion'])

        self.assertEqual('Legion', span.data['service'])

        self.assertEqual('aws:api.gateway', span.data['lambda']['trigger'])
        self.assertEqual('POST', span.data['http']['method'])
        self.assertEqual('/path/to/resource', span.data['http']['url'])
        self.assertEqual('/{proxy+}', span.data['http']['path_tpl'])
        if sys.version[:3] == '2.7':
            self.assertEqual(u"foo=[u'bar']", span.data['http']['params'])
        else:
            self.assertEqual("foo=['bar']", span.data['http']['params'])

    def test_api_gateway_trigger_tracing(self):
        with open(self.pwd + '/../data/lambda/api_gateway_event.json', 'r') as json_file:
            event = json.load(json_file)

        self.create_agent_and_setup_tracer()

        # Call the Instana Lambda Handler as we do in the real world.  It will initiate tracing and then
        # figure out the original (the users') Lambda Handler and execute it.
        # The original Lambda handler is set in os.environ["LAMBDA_HANDLER"]
        result = lambda_handler(event, self.context)

        assert isinstance(result, dict)
        assert 'headers' in result
        assert 'Server-Timing' in result['headers']

        time.sleep(1)
        payload = self.agent.collector.prepare_payload()

        self.assertTrue("metrics" in payload)
        self.assertTrue("spans" in payload)
        self.assertEqual(2, len(payload.keys()))

        self.assertTrue(isinstance(payload['metrics']['plugins'], list))
        self.assertTrue(len(payload['metrics']['plugins']) == 1)
        plugin_data = payload['metrics']['plugins'][0]

        self.assertEqual('com.instana.plugin.aws.lambda', plugin_data['name'])
        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', plugin_data['entityId'])

        self.assertEqual(1, len(payload['spans']))

        span = payload['spans'][0]
        self.assertEqual('aws.lambda.entry', span.n)
        self.assertEqual('d5cb361b256413a9', span.t)
        self.assertIsNotNone(span.s)
        self.assertEqual('0901d8ae4fbf1529', span.p)
        self.assertIsNotNone(span.ts)
        self.assertIsNotNone(span.d)

        server_timing_value = "intid;desc=%s" % span.t
        assert result['headers']['Server-Timing'] == server_timing_value

        self.assertEqual({'hl': True, 'cp': 'aws', 'e': 'arn:aws:lambda:us-east-2:12345:function:TestPython:1'},
                         span.f)

        self.assertTrue(span.sy)

        self.assertIsNone(span.ec)
        self.assertIsNone(span.data['lambda']['error'])

        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', span.data['lambda']['arn'])
        self.assertEqual(None, span.data['lambda']['alias'])
        self.assertEqual('python', span.data['lambda']['runtime'])
        self.assertEqual('TestPython', span.data['lambda']['functionName'])
        self.assertEqual('1', span.data['lambda']['functionVersion'])
        self.assertIsNone(span.data['service'])

        self.assertEqual('aws:api.gateway', span.data['lambda']['trigger'])
        self.assertEqual('POST', span.data['http']['method'])
        self.assertEqual('/path/to/resource', span.data['http']['url'])
        self.assertEqual('/{proxy+}', span.data['http']['path_tpl'])
        if sys.version[:3] == '2.7':
            self.assertEqual(u"foo=[u'bar']", span.data['http']['params'])
        else:
            self.assertEqual("foo=['bar']", span.data['http']['params'])

    def test_api_gateway_v2_trigger_tracing(self):
        with open(self.pwd + '/../data/lambda/api_gateway_v2_event.json', 'r') as json_file:
            event = json.load(json_file)

        self.create_agent_and_setup_tracer()

        # Call the Instana Lambda Handler as we do in the real world.  It will initiate tracing and then
        # figure out the original (the users') Lambda Handler and execute it.
        # The original Lambda handler is set in os.environ["LAMBDA_HANDLER"]
        result = lambda_handler(event, self.context)

        assert isinstance(result, dict)
        assert 'headers' in result
        assert 'Server-Timing' in result['headers']

        time.sleep(1)
        payload = self.agent.collector.prepare_payload()

        self.assertTrue("metrics" in payload)
        self.assertTrue("spans" in payload)
        self.assertEqual(2, len(payload.keys()))

        self.assertTrue(isinstance(payload['metrics']['plugins'], list))
        self.assertTrue(len(payload['metrics']['plugins']) == 1)
        plugin_data = payload['metrics']['plugins'][0]

        self.assertEqual('com.instana.plugin.aws.lambda', plugin_data['name'])
        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', plugin_data['entityId'])

        self.assertEqual(1, len(payload['spans']))

        span = payload['spans'][0]
        self.assertEqual('aws.lambda.entry', span.n)
        self.assertEqual('0000000000001234', span.t)
        self.assertIsNotNone(span.s)
        self.assertEqual('0000000000004567', span.p)
        self.assertIsNotNone(span.ts)
        self.assertIsNotNone(span.d)

        server_timing_value = "intid;desc=%s" % span.t
        assert result['headers']['Server-Timing'] == server_timing_value

        self.assertEqual({'hl': True, 'cp': 'aws', 'e': 'arn:aws:lambda:us-east-2:12345:function:TestPython:1'},
                         span.f)

        self.assertTrue(span.sy)

        self.assertIsNone(span.ec)
        self.assertIsNone(span.data['lambda']['error'])

        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', span.data['lambda']['arn'])
        self.assertEqual(None, span.data['lambda']['alias'])
        self.assertEqual('python', span.data['lambda']['runtime'])
        self.assertEqual('TestPython', span.data['lambda']['functionName'])
        self.assertEqual('1', span.data['lambda']['functionVersion'])
        self.assertIsNone(span.data['service'])

        self.assertEqual('aws:api.gateway', span.data['lambda']['trigger'])
        self.assertEqual('POST', span.data['http']['method'])
        self.assertEqual('/my/path', span.data['http']['url'])
        self.assertEqual('/my/{resource}', span.data['http']['path_tpl'])
        if sys.version[:3] == '2.7':
            self.assertEqual(u"q=term&secret=key", span.data['http']['params'])
        else:
            self.assertEqual("secret=key&q=term", span.data['http']['params'])


    def test_application_lb_trigger_tracing(self):
        with open(self.pwd + '/../data/lambda/api_gateway_event.json', 'r') as json_file:
            event = json.load(json_file)

        self.create_agent_and_setup_tracer()

        # Call the Instana Lambda Handler as we do in the real world.  It will initiate tracing and then
        # figure out the original (the users') Lambda Handler and execute it.
        # The original Lambda handler is set in os.environ["LAMBDA_HANDLER"]
        result = lambda_handler(event, self.context)

        assert isinstance(result, dict)
        assert 'headers' in result
        assert 'Server-Timing' in result['headers']

        time.sleep(1)
        payload = self.agent.collector.prepare_payload()

        self.assertTrue("metrics" in payload)
        self.assertTrue("spans" in payload)
        self.assertEqual(2, len(payload.keys()))

        self.assertTrue(isinstance(payload['metrics']['plugins'], list))
        self.assertTrue(len(payload['metrics']['plugins']) == 1)
        plugin_data = payload['metrics']['plugins'][0]

        self.assertEqual('com.instana.plugin.aws.lambda', plugin_data['name'])
        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', plugin_data['entityId'])

        self.assertEqual(1, len(payload['spans']))

        span = payload['spans'][0]
        self.assertEqual('aws.lambda.entry', span.n)
        self.assertEqual('d5cb361b256413a9', span.t)
        self.assertIsNotNone(span.s)
        self.assertEqual('0901d8ae4fbf1529', span.p)
        self.assertIsNotNone(span.ts)
        self.assertIsNotNone(span.d)

        server_timing_value = "intid;desc=%s" % span.t
        assert result['headers']['Server-Timing'] == server_timing_value

        self.assertEqual({'hl': True, 'cp': 'aws', 'e': 'arn:aws:lambda:us-east-2:12345:function:TestPython:1'},
                         span.f)

        self.assertTrue(span.sy)

        self.assertIsNone(span.ec)
        self.assertIsNone(span.data['lambda']['error'])

        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', span.data['lambda']['arn'])
        self.assertEqual(None, span.data['lambda']['alias'])
        self.assertEqual('python', span.data['lambda']['runtime'])
        self.assertEqual('TestPython', span.data['lambda']['functionName'])
        self.assertEqual('1', span.data['lambda']['functionVersion'])
        self.assertIsNone(span.data['service'])

        self.assertEqual('aws:api.gateway', span.data['lambda']['trigger'])
        self.assertEqual('POST', span.data['http']['method'])
        self.assertEqual('/path/to/resource', span.data['http']['url'])
        if sys.version[:3] == '2.7':
            self.assertEqual(u"foo=[u'bar']", span.data['http']['params'])
        else:
            self.assertEqual("foo=['bar']", span.data['http']['params'])

    def test_cloudwatch_trigger_tracing(self):
        with open(self.pwd + '/../data/lambda/cloudwatch_event.json', 'r') as json_file:
            event = json.load(json_file)

        self.create_agent_and_setup_tracer()

        # Call the Instana Lambda Handler as we do in the real world.  It will initiate tracing and then
        # figure out the original (the users') Lambda Handler and execute it.
        # The original Lambda handler is set in os.environ["LAMBDA_HANDLER"]
        result = lambda_handler(event, self.context)

        assert isinstance(result, dict)
        assert 'headers' in result
        assert 'Server-Timing' in result['headers']

        time.sleep(1)
        payload = self.agent.collector.prepare_payload()

        self.assertTrue("metrics" in payload)
        self.assertTrue("spans" in payload)
        self.assertEqual(2, len(payload.keys()))

        self.assertTrue(isinstance(payload['metrics']['plugins'], list))
        self.assertTrue(len(payload['metrics']['plugins']) == 1)
        plugin_data = payload['metrics']['plugins'][0]

        self.assertEqual('com.instana.plugin.aws.lambda', plugin_data['name'])
        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', plugin_data['entityId'])

        self.assertEqual(1, len(payload['spans']))

        span = payload['spans'][0]
        self.assertEqual('aws.lambda.entry', span.n)
        self.assertIsNotNone(span.t)
        self.assertIsNotNone(span.s)
        self.assertIsNone(span.p)
        self.assertIsNotNone(span.ts)
        self.assertIsNotNone(span.d)

        server_timing_value = "intid;desc=%s" % span.t
        assert result['headers']['Server-Timing'] == server_timing_value

        self.assertEqual({'hl': True, 'cp': 'aws', 'e': 'arn:aws:lambda:us-east-2:12345:function:TestPython:1'},
                         span.f)

        self.assertIsNone(span.sy)

        self.assertIsNone(span.ec)
        self.assertIsNone(span.data['lambda']['error'])

        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', span.data['lambda']['arn'])
        self.assertEqual(None, span.data['lambda']['alias'])
        self.assertEqual('python', span.data['lambda']['runtime'])
        self.assertEqual('TestPython', span.data['lambda']['functionName'])
        self.assertEqual('1', span.data['lambda']['functionVersion'])
        self.assertIsNone(span.data['service'])

        self.assertEqual('aws:cloudwatch.events', span.data['lambda']['trigger'])
        self.assertEqual('cdc73f9d-aea9-11e3-9d5a-835b769c0d9c', span.data["lambda"]["cw"]["events"]["id"])
        self.assertEqual(False, span.data["lambda"]["cw"]["events"]["more"])
        self.assertTrue(isinstance(span.data["lambda"]["cw"]["events"]["resources"], list))
        self.assertEqual(1, len(span.data["lambda"]["cw"]["events"]["resources"]))
        self.assertEqual('arn:aws:events:eu-west-1:123456789012:rule/ExampleRule',
                         span.data["lambda"]["cw"]["events"]["resources"][0])

    def test_cloudwatch_logs_trigger_tracing(self):
        with open(self.pwd + '/../data/lambda/cloudwatch_logs_event.json', 'r') as json_file:
            event = json.load(json_file)

        self.create_agent_and_setup_tracer()

        # Call the Instana Lambda Handler as we do in the real world.  It will initiate tracing and then
        # figure out the original (the users') Lambda Handler and execute it.
        # The original Lambda handler is set in os.environ["LAMBDA_HANDLER"]
        result = lambda_handler(event, self.context)

        assert isinstance(result, dict)
        assert 'headers' in result
        assert 'Server-Timing' in result['headers']

        time.sleep(1)
        payload = self.agent.collector.prepare_payload()

        self.assertTrue("metrics" in payload)
        self.assertTrue("spans" in payload)
        self.assertEqual(2, len(payload.keys()))

        self.assertTrue(isinstance(payload['metrics']['plugins'], list))
        self.assertTrue(len(payload['metrics']['plugins']) == 1)
        plugin_data = payload['metrics']['plugins'][0]

        self.assertEqual('com.instana.plugin.aws.lambda', plugin_data['name'])
        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', plugin_data['entityId'])

        self.assertEqual(1, len(payload['spans']))

        span = payload['spans'][0]
        self.assertEqual('aws.lambda.entry', span.n)
        self.assertIsNotNone(span.t)
        self.assertIsNotNone(span.s)
        self.assertIsNone(span.p)
        self.assertIsNotNone(span.ts)
        self.assertIsNotNone(span.d)

        server_timing_value = "intid;desc=%s" % span.t
        assert result['headers']['Server-Timing'] == server_timing_value

        self.assertEqual({'hl': True, 'cp': 'aws', 'e': 'arn:aws:lambda:us-east-2:12345:function:TestPython:1'},
                         span.f)

        self.assertIsNone(span.sy)

        self.assertIsNone(span.ec)
        self.assertIsNone(span.data['lambda']['error'])

        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', span.data['lambda']['arn'])
        self.assertEqual(None, span.data['lambda']['alias'])
        self.assertEqual('python', span.data['lambda']['runtime'])
        self.assertEqual('TestPython', span.data['lambda']['functionName'])
        self.assertEqual('1', span.data['lambda']['functionVersion'])
        self.assertIsNone(span.data['service'])

        self.assertEqual('aws:cloudwatch.logs', span.data['lambda']['trigger'])
        self.assertFalse("decodingError" in span.data['lambda']['cw']['logs'])
        self.assertEqual('testLogGroup', span.data['lambda']['cw']['logs']['group'])
        self.assertEqual('testLogStream', span.data['lambda']['cw']['logs']['stream'])
        self.assertEqual(None, span.data['lambda']['cw']['logs']['more'])
        self.assertTrue(isinstance(span.data['lambda']['cw']['logs']['events'], list))
        self.assertEqual(2, len(span.data['lambda']['cw']['logs']['events']))
        self.assertEqual('[ERROR] First test message', span.data['lambda']['cw']['logs']['events'][0])
        self.assertEqual('[ERROR] Second test message', span.data['lambda']['cw']['logs']['events'][1])

    def test_s3_trigger_tracing(self):
        with open(self.pwd + '/../data/lambda/s3_event.json', 'r') as json_file:
            event = json.load(json_file)

        self.create_agent_and_setup_tracer()

        # Call the Instana Lambda Handler as we do in the real world.  It will initiate tracing and then
        # figure out the original (the users') Lambda Handler and execute it.
        # The original Lambda handler is set in os.environ["LAMBDA_HANDLER"]
        result = lambda_handler(event, self.context)

        assert isinstance(result, dict)
        assert 'headers' in result
        assert 'Server-Timing' in result['headers']

        time.sleep(1)
        payload = self.agent.collector.prepare_payload()

        self.assertTrue("metrics" in payload)
        self.assertTrue("spans" in payload)
        self.assertEqual(2, len(payload.keys()))

        self.assertTrue(isinstance(payload['metrics']['plugins'], list))
        self.assertTrue(len(payload['metrics']['plugins']) == 1)
        plugin_data = payload['metrics']['plugins'][0]

        self.assertEqual('com.instana.plugin.aws.lambda', plugin_data['name'])
        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', plugin_data['entityId'])

        self.assertEqual(1, len(payload['spans']))

        span = payload['spans'][0]
        self.assertEqual('aws.lambda.entry', span.n)
        self.assertIsNotNone(span.t)
        self.assertIsNotNone(span.s)
        self.assertIsNone(span.p)
        self.assertIsNotNone(span.ts)
        self.assertIsNotNone(span.d)

        server_timing_value = "intid;desc=%s" % span.t
        assert result['headers']['Server-Timing'] == server_timing_value

        self.assertEqual({'hl': True, 'cp': 'aws', 'e': 'arn:aws:lambda:us-east-2:12345:function:TestPython:1'},
                         span.f)

        self.assertIsNone(span.sy)

        self.assertIsNone(span.ec)
        self.assertIsNone(span.data['lambda']['error'])

        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', span.data['lambda']['arn'])
        self.assertEqual(None, span.data['lambda']['alias'])
        self.assertEqual('python', span.data['lambda']['runtime'])
        self.assertEqual('TestPython', span.data['lambda']['functionName'])
        self.assertEqual('1', span.data['lambda']['functionVersion'])
        self.assertIsNone(span.data['service'])

        self.assertEqual('aws:s3', span.data['lambda']['trigger'])
        self.assertTrue(isinstance(span.data["lambda"]["s3"]["events"], list))
        events = span.data["lambda"]["s3"]["events"]
        self.assertEqual(1, len(events))
        event = events[0]
        self.assertEqual('ObjectCreated:Put', event['event'])
        self.assertEqual('example-bucket', event['bucket'])
        self.assertEqual('test/key', event['object'])

    def test_sqs_trigger_tracing(self):
        with open(self.pwd + '/../data/lambda/sqs_event.json', 'r') as json_file:
            event = json.load(json_file)

        self.create_agent_and_setup_tracer()

        # Call the Instana Lambda Handler as we do in the real world.  It will initiate tracing and then
        # figure out the original (the users') Lambda Handler and execute it.
        # The original Lambda handler is set in os.environ["LAMBDA_HANDLER"]
        result = lambda_handler(event, self.context)

        assert isinstance(result, dict)
        assert 'headers' in result
        assert 'Server-Timing' in result['headers']

        time.sleep(1)
        payload = self.agent.collector.prepare_payload()

        self.assertTrue("metrics" in payload)
        self.assertTrue("spans" in payload)
        self.assertEqual(2, len(payload.keys()))

        self.assertTrue(isinstance(payload['metrics']['plugins'], list))
        self.assertTrue(len(payload['metrics']['plugins']) == 1)
        plugin_data = payload['metrics']['plugins'][0]

        self.assertEqual('com.instana.plugin.aws.lambda', plugin_data['name'])
        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', plugin_data['entityId'])

        self.assertEqual(1, len(payload['spans']))

        span = payload['spans'][0]
        self.assertEqual('aws.lambda.entry', span.n)
        self.assertIsNotNone(span.t)
        self.assertIsNotNone(span.s)
        self.assertIsNone(span.p)
        self.assertIsNotNone(span.ts)
        self.assertIsNotNone(span.d)

        server_timing_value = "intid;desc=%s" % span.t
        assert result['headers']['Server-Timing'] == server_timing_value

        self.assertEqual({'hl': True, 'cp': 'aws', 'e': 'arn:aws:lambda:us-east-2:12345:function:TestPython:1'},
                         span.f)

        self.assertIsNone(span.sy)

        self.assertIsNone(span.ec)
        self.assertIsNone(span.data['lambda']['error'])

        self.assertEqual('arn:aws:lambda:us-east-2:12345:function:TestPython:1', span.data['lambda']['arn'])
        self.assertEqual(None, span.data['lambda']['alias'])
        self.assertEqual('python', span.data['lambda']['runtime'])
        self.assertEqual('TestPython', span.data['lambda']['functionName'])
        self.assertEqual('1', span.data['lambda']['functionVersion'])
        self.assertIsNone(span.data['service'])

        self.assertEqual('aws:sqs', span.data['lambda']['trigger'])
        self.assertTrue(isinstance(span.data["lambda"]["sqs"]["messages"], list))
        messages = span.data["lambda"]["sqs"]["messages"]
        self.assertEqual(1, len(messages))
        message = messages[0]
        self.assertEqual('arn:aws:sqs:us-west-1:123456789012:MyQueue', message['queue'])

    def test_read_query_params(self):
        event = { "queryStringParameters": {"foo": "bar" },
                  "multiValueQueryStringParameters": { "foo": ["bar"] } }
        params = read_http_query_params(event)
        self.assertEqual("foo=['bar']", params)

    def test_read_query_params_with_none_data(self):
        event = { "queryStringParameters": None,
                  "multiValueQueryStringParameters": None }
        params = read_http_query_params(event)
        self.assertEqual("", params)

    def test_read_query_params_with_bad_event(self):
        event = None
        params = read_http_query_params(event)
        self.assertEqual("", params)

    def test_arn_parsing(self):
        ctx = MockContext()

        assert(normalize_aws_lambda_arn(ctx) == "arn:aws:lambda:us-east-2:12345:function:TestPython:1")

        # Without version should return a fully qualified ARN (with version)
        ctx.invoked_function_arn = "arn:aws:lambda:us-east-2:12345:function:TestPython"
        assert(normalize_aws_lambda_arn(ctx) == "arn:aws:lambda:us-east-2:12345:function:TestPython:1")

        # Fully qualified already with the '$LATEST' special tag
        ctx.invoked_function_arn = "arn:aws:lambda:us-east-2:12345:function:TestPython:$LATEST"
        assert(normalize_aws_lambda_arn(ctx) == "arn:aws:lambda:us-east-2:12345:function:TestPython:$LATEST")

    def test_agent_default_log_level(self):
        self.create_agent_and_setup_tracer()
        assert self.agent.options.log_level == logging.WARNING

    def test_agent_custom_log_level(self):
        os.environ['INSTANA_LOG_LEVEL'] = "eRror"
        self.create_agent_and_setup_tracer()
        assert self.agent.options.log_level == logging.ERROR



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from ..singletons import agent, tracer, async_tracer, tornado_tracer
from ..log import logger


def extract_custom_headers(tracing_scope, headers):
    try:
        for custom_header in agent.options.extra_http_headers:
            # Headers are in the following format: b'x-header-1'
            for header_key, value in headers.items():
                if header_key.lower() == custom_header.lower():
                    tracing_scope.span.set_tag("http.header.%s" % custom_header, value)
    except Exception:
        logger.debug("extract_custom_headers: ", exc_info=True)


def get_active_tracer():
    try:
        if tracer.active_span:
            return tracer
        elif async_tracer.active_span:
            return async_tracer
        elif tornado_tracer.active_span:
            return tornado_tracer
        else:
            return None
    except Exception:
        # Do not try to log this with instana, as there is no active tracer and there will be an infinite loop at least
        # for PY2
        return None



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import re
import os
import sys

from ..log import logger

def get_py_source(filename):
    """
    Retrieves and returns the source code for any Python
    files requested by the UI via the host agent

    @param filename [String] The fully qualified path to a file
    """
    response = None
    try:
        if regexp_py.search(filename) is None:
            response = {"error": "Only Python source files are allowed. (*.py)"}
        else:
            pysource = ""
            with open(filename, 'r') as pyfile:
                pysource = pyfile.read()

            response = {"data": pysource}

    except Exception as exc:
        response = {"error": str(exc)}

    return response


# Used by get_py_source
regexp_py = re.compile(r"\.py$")


def determine_service_name():
    """ This function makes a best effort to name this application process. """

    # One environment variable to rule them all
    if "INSTANA_SERVICE_NAME" in os.environ:
        return os.environ["INSTANA_SERVICE_NAME"]

    # Now best effort in naming this process.  No nice package.json like in Node.js
    # so we do best effort detection here.
    app_name = "python"  # the default name
    basename = None

    try:
        if not hasattr(sys, 'argv'):
            proc_cmdline = get_proc_cmdline(as_string=False)
            return os.path.basename(proc_cmdline[0])

        # Get first argument that is not an CLI option
        for candidate in sys.argv:
            if len(candidate) > 0 and candidate[0] != '-':
                basename = candidate
                break

        # If nothing found, fall back to executable
        if basename is None:
            basename = os.path.basename(sys.executable)
        else:
            # Assure leading paths are stripped
            basename = os.path.basename(basename)

        if basename == "gunicorn":
            if 'setproctitle' in sys.modules:
                # With the setproctitle package, gunicorn renames their processes
                # to pretty things - we use those by default
                # gunicorn: master [djface.wsgi]
                # gunicorn: worker [djface.wsgi]
                app_name = get_proc_cmdline(as_string=True)
            else:
                app_name = basename
        elif "FLASK_APP" in os.environ:
            app_name = os.environ["FLASK_APP"]
        elif "DJANGO_SETTINGS_MODULE" in os.environ:
            app_name = os.environ["DJANGO_SETTINGS_MODULE"].split('.')[0]
        elif basename == '':
            if sys.stdout.isatty():
                app_name = "Interactive Console"
            else:
                # No arguments.  Take executable as app_name
                app_name = os.path.basename(sys.executable)
        else:
            # Last chance.  app_name for "python main.py" would be "main.py" here.
            app_name = basename

        # We should have a good app_name by this point.
        # Last conditional, if uwsgi, then wrap the name
        # with the uwsgi process type
        if basename == "uwsgi":
            # We have an app name by this point.  Now if running under
            # uwsgi, augment the app name
            try:
                import uwsgi

                if app_name == "uwsgi":
                    app_name = ""
                else:
                    app_name = " [%s]" % app_name

                if os.getpid() == uwsgi.masterpid():
                    uwsgi_type = "uWSGI master%s"
                else:
                    uwsgi_type = "uWSGI worker%s"

                app_name = uwsgi_type % app_name
            except ImportError:
                pass
    except Exception:
        logger.debug("non-fatal get_application_name: ", exc_info=True)

    return app_name

def get_proc_cmdline(as_string=False):
    """
    Parse the proc file system for the command line of this process.  If not available, then return a default.
    Return is dependent on the value of `as_string`.  If True, return the full command line as a string,
    otherwise a list.
    """
    name = "python"
    if os.path.isfile("/proc/self/cmdline"):
        with open("/proc/self/cmdline") as cmd:
            name = cmd.read()
    else:
        # Most likely not on a *nix based OS.  Return a default
        if as_string is True:
            return name
        else:
            return [name]

    # /proc/self/command line will have strings with null bytes such as "/usr/bin/python\0-s\0-d\0".  This
    # bit will prep the return value and drop the trailing null byte
    parts = name.split('\0')
    parts.pop()

    if as_string is True:
        parts = " ".join(parts)

    return parts


# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from __future__ import absolute_import

import time
import pytest
import requests
import multiprocessing
from instana.singletons import tracer
from ..helpers import testenv
from ..helpers import get_first_span_by_filter
from ..test_utils import _TraceContextMixin
import sys
import unittest


@pytest.mark.skipif(sys.version_info[0] < 3 or (sys.version_info[0] == 3 and sys.version_info[1] < 7),
                    reason="testing sanic for python 3.7 and up")
class TestSanic(unittest.TestCase, _TraceContextMixin):

    def setUp(self):
        from tests.apps.sanic_app import launch_sanic
        self.proc = multiprocessing.Process(target=launch_sanic, args=(), daemon=True)
        self.proc.start()
        time.sleep(2)

    def tearDown(self):
        self.proc.kill()

    def test_vanilla_get(self):
        result = requests.get(testenv["sanic_server"] + '/')

        self.assertEqual(result.status_code, 200)
        self.assertIn("X-INSTANA-T", result.headers)
        self.assertIn("X-INSTANA-S", result.headers)
        self.assertIn("X-INSTANA-L", result.headers)
        self.assertEqual(result.headers["X-INSTANA-L"], "1")
        self.assertIn("Server-Timing", result.headers)
        spans = tracer.recorder.queued_spans()
        self.assertEqual(len(spans), 1)
        self.assertEqual(spans[0].n, 'asgi')

    def test_basic_get(self):
        result = None
        with tracer.start_active_span('test'):
            result = requests.get(testenv["sanic_server"] + '/')

        self.assertEqual(result.status_code, 200)

        spans = tracer.recorder.queued_spans()
        self.assertEqual(len(spans), 3)

        span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
        test_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(test_span)

        span_filter = lambda span: span.n == "urllib3"
        urllib3_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(urllib3_span)

        span_filter = lambda span: span.n == 'asgi'
        asgi_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(asgi_span)

        self.assertTraceContextPropagated(test_span, urllib3_span)
        self.assertTraceContextPropagated(urllib3_span, asgi_span)

        self.assertIn("X-INSTANA-T", result.headers)
        self.assertEqual(result.headers["X-INSTANA-T"], asgi_span.t)
        self.assertIn("X-INSTANA-S", result.headers)
        self.assertEqual(result.headers["X-INSTANA-S"], asgi_span.s)
        self.assertIn("X-INSTANA-L", result.headers)
        self.assertEqual(result.headers["X-INSTANA-L"], '1')
        self.assertIn("Server-Timing", result.headers)
        self.assertEqual(result.headers["Server-Timing"], ("intid;desc=%s" % asgi_span.t))

        self.assertIsNone(asgi_span.ec)
        assert (asgi_span.data['http']['host'] == '127.0.0.1:1337')
        assert (asgi_span.data['http']['path'] == '/')
        assert (asgi_span.data['http']['path_tpl'] == '/')
        assert (asgi_span.data['http']['method'] == 'GET')
        assert (asgi_span.data['http']['status'] == 200)
        assert (asgi_span.data['http']['error'] is None)
        assert (asgi_span.data['http']['params'] is None)

    def test_404(self):
        result = None
        with tracer.start_active_span('test'):
            result = requests.get(testenv["sanic_server"] + '/foo/not_an_int')

        self.assertEqual(result.status_code, 404)

        spans = tracer.recorder.queued_spans()
        self.assertEqual(len(spans), 3)

        span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
        test_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(test_span)

        span_filter = lambda span: span.n == "urllib3"
        urllib3_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(urllib3_span)

        span_filter = lambda span: span.n == 'asgi'
        asgi_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(asgi_span)

        self.assertTraceContextPropagated(test_span, urllib3_span)
        self.assertTraceContextPropagated(urllib3_span, asgi_span)

        self.assertIn("X-INSTANA-T", result.headers)
        self.assertEqual(result.headers["X-INSTANA-T"], asgi_span.t)
        self.assertIn("X-INSTANA-S", result.headers)
        self.assertEqual(result.headers["X-INSTANA-S"], asgi_span.s)
        self.assertIn("X-INSTANA-L", result.headers)
        self.assertEqual(result.headers["X-INSTANA-L"], '1')
        self.assertIn("Server-Timing", result.headers)
        self.assertEqual(result.headers["Server-Timing"], ("intid;desc=%s" % asgi_span.t))

        self.assertIsNone(asgi_span.ec)
        assert (asgi_span.data['http']['host'] == '127.0.0.1:1337')
        assert (asgi_span.data['http']['path'] == '/foo/not_an_int')
        assert (asgi_span.data['http']['path_tpl'] is None)
        assert (asgi_span.data['http']['method'] == 'GET')
        assert (asgi_span.data['http']['status'] == 404)
        assert (asgi_span.data['http']['error'] is None)
        assert (asgi_span.data['http']['params'] is None)

    def test_sanic_exception(self):
        result = None
        with tracer.start_active_span('test'):
            result = requests.get(testenv["sanic_server"] + '/wrong')

        self.assertEqual(result.status_code, 400)

        spans = tracer.recorder.queued_spans()
        self.assertEqual(len(spans), 3)

        span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
        test_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(test_span)

        span_filter = lambda span: span.n == "urllib3"
        urllib3_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(urllib3_span)

        span_filter = lambda span: span.n == 'asgi'
        asgi_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(asgi_span)

        self.assertTraceContextPropagated(test_span, urllib3_span)
        self.assertTraceContextPropagated(urllib3_span, asgi_span)

        self.assertIn("X-INSTANA-T", result.headers)
        self.assertEqual(result.headers["X-INSTANA-T"], asgi_span.t)
        self.assertIn("X-INSTANA-S", result.headers)
        self.assertEqual(result.headers["X-INSTANA-S"], asgi_span.s)
        self.assertIn("X-INSTANA-L", result.headers)
        self.assertEqual(result.headers["X-INSTANA-L"], '1')
        self.assertIn("Server-Timing", result.headers)
        self.assertEqual(result.headers["Server-Timing"], ("intid;desc=%s" % asgi_span.t))

        self.assertIsNone(asgi_span.ec)
        assert (asgi_span.data['http']['host'] == '127.0.0.1:1337')
        assert (asgi_span.data['http']['path'] == '/wrong')
        assert (asgi_span.data['http']['path_tpl'] == '/wrong')
        assert (asgi_span.data['http']['method'] == 'GET')
        assert (asgi_span.data['http']['status'] == 400)
        assert (asgi_span.data['http']['error'] is None)
        assert (asgi_span.data['http']['params'] is None)

    def test_500_instana_exception(self):
        result = None
        with tracer.start_active_span('test'):
            result = requests.get(testenv["sanic_server"] + '/instana_exception')

        self.assertEqual(result.status_code, 500)

        spans = tracer.recorder.queued_spans()
        self.assertEqual(len(spans), 4)

        span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
        test_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(test_span)

        span_filter = lambda span: span.n == "urllib3"
        urllib3_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(urllib3_span)

        span_filter = lambda span: span.n == 'asgi'
        asgi_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(asgi_span)

        self.assertTraceContextPropagated(test_span, urllib3_span)
        self.assertTraceContextPropagated(urllib3_span, asgi_span)

        self.assertIn("X-INSTANA-T", result.headers)
        self.assertEqual(result.headers["X-INSTANA-T"], asgi_span.t)
        self.assertIn("X-INSTANA-S", result.headers)
        self.assertEqual(result.headers["X-INSTANA-S"], asgi_span.s)
        self.assertIn("X-INSTANA-L", result.headers)
        self.assertEqual(result.headers["X-INSTANA-L"], '1')
        self.assertIn("Server-Timing", result.headers)
        self.assertEqual(result.headers["Server-Timing"], ("intid;desc=%s" % asgi_span.t))

        self.assertEqual(asgi_span.ec, 1)
        assert (asgi_span.data['http']['host'] == '127.0.0.1:1337')
        assert (asgi_span.data['http']['path'] == '/instana_exception')
        assert (asgi_span.data['http']['path_tpl'] == '/instana_exception')
        assert (asgi_span.data['http']['method'] == 'GET')
        assert (asgi_span.data['http']['status'] == 500)
        assert (asgi_span.data['http']['error'] is None)
        assert (asgi_span.data['http']['params'] is None)

    def test_500(self):
        result = None
        with tracer.start_active_span('test'):
            result = requests.get(testenv["sanic_server"] + '/test_request_args')

        self.assertEqual(result.status_code, 500)

        spans = tracer.recorder.queued_spans()
        self.assertEqual(len(spans), 4)

        span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
        test_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(test_span)

        span_filter = lambda span: span.n == "urllib3"
        urllib3_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(urllib3_span)

        span_filter = lambda span: span.n == 'asgi'
        asgi_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(asgi_span)

        self.assertTraceContextPropagated(test_span, urllib3_span)
        self.assertTraceContextPropagated(urllib3_span, asgi_span)

        self.assertIn("X-INSTANA-T", result.headers)
        self.assertEqual(result.headers["X-INSTANA-T"], asgi_span.t)
        self.assertIn("X-INSTANA-S", result.headers)
        self.assertEqual(result.headers["X-INSTANA-S"], asgi_span.s)
        self.assertIn("X-INSTANA-L", result.headers)
        self.assertEqual(result.headers["X-INSTANA-L"], '1')
        self.assertIn("Server-Timing", result.headers)
        self.assertEqual(result.headers["Server-Timing"], ("intid;desc=%s" % asgi_span.t))

        self.assertEqual(asgi_span.ec, 1)
        assert (asgi_span.data['http']['host'] == '127.0.0.1:1337')
        assert (asgi_span.data['http']['path'] == '/test_request_args')
        assert (asgi_span.data['http']['path_tpl'] == '/test_request_args')
        assert (asgi_span.data['http']['method'] == 'GET')
        assert (asgi_span.data['http']['status'] == 500)
        assert (asgi_span.data['http']['error'] == 'Something went wrong.')
        assert (asgi_span.data['http']['params'] is None)

    def test_path_templates(self):
        result = None
        with tracer.start_active_span('test'):
            result = requests.get(testenv["sanic_server"] + '/foo/1')

        self.assertEqual(result.status_code, 200)

        spans = tracer.recorder.queued_spans()
        self.assertEqual(len(spans), 3)

        span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
        test_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(test_span)

        span_filter = lambda span: span.n == "urllib3"
        urllib3_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(urllib3_span)

        span_filter = lambda span: span.n == 'asgi'
        asgi_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(asgi_span)

        self.assertTraceContextPropagated(test_span, urllib3_span)
        self.assertTraceContextPropagated(urllib3_span, asgi_span)

        self.assertIn("X-INSTANA-T", result.headers)
        self.assertEqual(result.headers["X-INSTANA-T"], asgi_span.t)
        self.assertIn("X-INSTANA-S", result.headers)
        self.assertEqual(result.headers["X-INSTANA-S"], asgi_span.s)
        self.assertIn("X-INSTANA-L", result.headers)
        self.assertEqual(result.headers["X-INSTANA-L"], '1')
        self.assertIn("Server-Timing", result.headers)
        self.assertEqual(result.headers["Server-Timing"], ("intid;desc=%s" % asgi_span.t))

        self.assertIsNone(asgi_span.ec)
        assert (asgi_span.data['http']['host'] == '127.0.0.1:1337')
        assert (asgi_span.data['http']['path'] == '/foo/1')
        assert (asgi_span.data['http']['path_tpl'] == '/foo/<foo_id:int>')
        assert (asgi_span.data['http']['method'] == 'GET')
        assert (asgi_span.data['http']['status'] == 200)
        assert (asgi_span.data['http']['error'] is None)
        assert (asgi_span.data['http']['params'] is None)

    def test_secret_scrubbing(self):
        result = None
        with tracer.start_active_span('test'):
            result = requests.get(testenv["sanic_server"] + '/?secret=shhh')

        self.assertEqual(result.status_code, 200)

        spans = tracer.recorder.queued_spans()
        self.assertEqual(len(spans), 3)

        span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
        test_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(test_span)

        span_filter = lambda span: span.n == "urllib3"
        urllib3_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(urllib3_span)

        span_filter = lambda span: span.n == 'asgi'
        asgi_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(asgi_span)

        self.assertTraceContextPropagated(test_span, urllib3_span)
        self.assertTraceContextPropagated(urllib3_span, asgi_span)

        self.assertIn("X-INSTANA-T", result.headers)
        self.assertEqual(result.headers["X-INSTANA-T"], asgi_span.t)
        self.assertIn("X-INSTANA-S", result.headers)
        self.assertEqual(result.headers["X-INSTANA-S"], asgi_span.s)
        self.assertIn("X-INSTANA-L", result.headers)
        self.assertEqual(result.headers["X-INSTANA-L"], '1')
        self.assertIn("Server-Timing", result.headers)
        self.assertEqual(result.headers["Server-Timing"], ("intid;desc=%s" % asgi_span.t))

        self.assertIsNone(asgi_span.ec)
        assert (asgi_span.data['http']['host'] == '127.0.0.1:1337')
        assert (asgi_span.data['http']['path'] == '/')
        assert (asgi_span.data['http']['path_tpl'] == '/')
        assert (asgi_span.data['http']['method'] == 'GET')
        assert (asgi_span.data['http']['status'] == 200)
        assert (asgi_span.data['http']['error'] is None)
        assert (asgi_span.data['http']['params'] == 'secret=<redacted>')

    def test_synthetic_request(self):
        request_headers = {
            'X-INSTANA-SYNTHETIC': '1'
        }
        with tracer.start_active_span('test'):
            result = requests.get(testenv["sanic_server"] + '/', headers=request_headers)

        self.assertEqual(result.status_code, 200)

        spans = tracer.recorder.queued_spans()
        self.assertEqual(len(spans), 3)

        span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
        test_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(test_span)

        span_filter = lambda span: span.n == "urllib3"
        urllib3_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(urllib3_span)

        span_filter = lambda span: span.n == 'asgi'
        asgi_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(asgi_span)

        self.assertTraceContextPropagated(test_span, urllib3_span)
        self.assertTraceContextPropagated(urllib3_span, asgi_span)

        self.assertIn("X-INSTANA-T", result.headers)
        self.assertEqual(result.headers["X-INSTANA-T"], asgi_span.t)
        self.assertIn("X-INSTANA-S", result.headers)
        self.assertEqual(result.headers["X-INSTANA-S"], asgi_span.s)
        self.assertIn("X-INSTANA-L", result.headers)
        self.assertEqual(result.headers["X-INSTANA-L"], '1')
        self.assertIn("Server-Timing", result.headers)
        self.assertEqual(result.headers["Server-Timing"], ("intid;desc=%s" % asgi_span.t))

        self.assertIsNone(asgi_span.ec)
        assert (asgi_span.data['http']['host'] == '127.0.0.1:1337')
        assert (asgi_span.data['http']['path'] == '/')
        assert (asgi_span.data['http']['path_tpl'] == '/')
        assert (asgi_span.data['http']['method'] == 'GET')
        assert (asgi_span.data['http']['status'] == 200)
        assert (asgi_span.data['http']['error'] is None)
        assert (asgi_span.data['http']['params'] is None)

        self.assertIsNotNone(asgi_span.sy)
        self.assertIsNone(urllib3_span.sy)
        self.assertIsNone(test_span.sy)

    def test_custom_header_capture(self):
        request_headers = {
            'X-Capture-This': 'this',
            'X-Capture-That': 'that'
        }
        with tracer.start_active_span('test'):
            result = requests.get(testenv["sanic_server"] + '/', headers=request_headers)

        self.assertEqual(result.status_code, 200)

        spans = tracer.recorder.queued_spans()
        self.assertEqual(len(spans), 3)

        span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
        test_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(test_span)

        span_filter = lambda span: span.n == "urllib3"
        urllib3_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(urllib3_span)

        span_filter = lambda span: span.n == 'asgi'
        asgi_span = get_first_span_by_filter(spans, span_filter)
        self.assertIsNotNone(asgi_span)

        self.assertTraceContextPropagated(test_span, urllib3_span)
        self.assertTraceContextPropagated(urllib3_span, asgi_span)

        self.assertIn("X-INSTANA-T", result.headers)
        self.assertEqual(result.headers["X-INSTANA-T"], asgi_span.t)
        self.assertIn("X-INSTANA-S", result.headers)
        self.assertEqual(result.headers["X-INSTANA-S"], asgi_span.s)
        self.assertIn("X-INSTANA-L", result.headers)
        self.assertEqual(result.headers["X-INSTANA-L"], '1')
        self.assertIn("Server-Timing", result.headers)
        self.assertEqual(result.headers["Server-Timing"], ("intid;desc=%s" % asgi_span.t))

        self.assertIsNone(asgi_span.ec)
        assert (asgi_span.data['http']['host'] == '127.0.0.1:1337')
        assert (asgi_span.data['http']['path'] == '/')
        assert (asgi_span.data['http']['path_tpl'] == '/')
        assert (asgi_span.data['http']['method'] == 'GET')
        assert (asgi_span.data['http']['status'] == 200)
        assert (asgi_span.data['http']['error'] is None)
        assert (asgi_span.data['http']['params'] is None)

        assert ("X-Capture-This" in asgi_span.data["http"]["header"])
        assert ("this" == asgi_span.data["http"]["header"]["X-Capture-This"])
        assert ("X-Capture-That" in asgi_span.data["http"]["header"])
        assert ("that" == asgi_span.data["http"]["header"]["X-Capture-That"])



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import sys

from instana.log import logger
from instana.util.ids import header_to_id, header_to_long_id
from instana.span_context import SpanContext
from instana.w3c_trace_context.traceparent import Traceparent
from instana.w3c_trace_context.tracestate import Tracestate
import os

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3


# The carrier can be a dict or a list.
# Using the trace header as an example, it can be in the following forms
# for extraction:
#   X-Instana-T
#   HTTP_X_INSTANA_T
#
# The second form above is found in places like Django middleware for
# incoming requests.
#
# For injection, we only support the standard format:
#   X-Instana-T


class BasePropagator(object):
    HEADER_KEY_T = 'X-INSTANA-T'
    HEADER_KEY_S = 'X-INSTANA-S'
    HEADER_KEY_L = 'X-INSTANA-L'
    HEADER_KEY_SYNTHETIC = 'X-INSTANA-SYNTHETIC'
    HEADER_KEY_TRACEPARENT = "traceparent"
    HEADER_KEY_TRACESTATE = "tracestate"

    LC_HEADER_KEY_T = 'x-instana-t'
    LC_HEADER_KEY_S = 'x-instana-s'
    LC_HEADER_KEY_L = 'x-instana-l'
    LC_HEADER_KEY_SYNTHETIC = 'x-instana-synthetic'

    ALT_LC_HEADER_KEY_T = 'http_x_instana_t'
    ALT_LC_HEADER_KEY_S = 'http_x_instana_s'
    ALT_LC_HEADER_KEY_L = 'http_x_instana_l'
    ALT_LC_HEADER_KEY_SYNTHETIC = 'http_x_instana_synthetic'
    ALT_HEADER_KEY_TRACEPARENT = "http_traceparent"
    ALT_HEADER_KEY_TRACESTATE = "http_tracestate"

    # ByteArray variations
    B_HEADER_KEY_T = b'x-instana-t'
    B_HEADER_KEY_S = b'x-instana-s'
    B_HEADER_KEY_L = b'x-instana-l'
    B_HEADER_KEY_SYNTHETIC = b'x-instana-synthetic'
    B_HEADER_SERVER_TIMING = b'server-timing'
    B_HEADER_KEY_TRACEPARENT = b'traceparent'
    B_HEADER_KEY_TRACESTATE = b'tracestate'

    B_ALT_LC_HEADER_KEY_T = b'http_x_instana_t'
    B_ALT_LC_HEADER_KEY_S = b'http_x_instana_s'
    B_ALT_LC_HEADER_KEY_L = b'http_x_instana_l'
    B_ALT_LC_HEADER_KEY_SYNTHETIC = b'http_x_instana_synthetic'
    B_ALT_HEADER_KEY_TRACEPARENT = b'http_traceparent'
    B_ALT_HEADER_KEY_TRACESTATE = b'http_tracestate'

    def __init__(self):
        self._tp = Traceparent()
        self._ts = Tracestate()

    @staticmethod
    def _extract_headers_dict(carrier):
        """
        This method converts the incoming carrier into a dict
        :param carrier:
        :return: dc dictionary
        """
        try:
            if isinstance(carrier, dict):
                dc = carrier
            elif hasattr(carrier, "__dict__"):
                dc = carrier.__dict__
            else:
                dc = dict(carrier)
        except Exception:
            logger.debug("extract: Couldn't convert %s", carrier)
            dc = None

        return dc

    @staticmethod
    def _get_ctx_level(level):
        """
        Extract the level value and return it, as it may include correlation values
        :param level:
        :return:
        """
        try:
            ctx_level = int(level.split(",")[0]) if level else 1
        except Exception:
            ctx_level = 1
        return ctx_level

    @staticmethod
    def _set_correlation_properties(level, ctx):
        """
        Set the correlation values if they are present
        :param level:
        :param ctx:
        :return:
        """
        try:
            ctx.correlation_type = level.split(",")[1].split("correlationType=")[1].split(";")[0]
            if "correlationId" in level:
                ctx.correlation_id = level.split(",")[1].split("correlationId=")[1].split(";")[0]
        except Exception:
            logger.debug("extract instana correlation type/id error:", exc_info=True)

    def _get_participating_trace_context(self, span_context):
        """
        This method is called for getting the updated traceparent and tracestate values
        :param span_context:
        :return: traceparent, tracestate
        """
        if span_context.long_trace_id and not span_context.trace_parent:
            tp_trace_id = span_context.long_trace_id
        else:
            tp_trace_id = span_context.trace_id
        traceparent = span_context.traceparent
        tracestate = span_context.tracestate
        traceparent = self._tp.update_traceparent(traceparent, tp_trace_id, span_context.span_id, span_context.level)
        tracestate = self._ts.update_tracestate(tracestate, span_context.trace_id, span_context.span_id)
        return traceparent, tracestate

    def __determine_span_context(self, trace_id, span_id, level, synthetic, traceparent, tracestate,
                                 disable_w3c_trace_context):
        """
        This method determines the span context depending on a set of conditions being met
        Detailed description of the conditions can be found here:
        https://github.com/instana/technical-documentation/tree/master/tracing/specification#http-processing-for-instana-tracers
        :param trace_id: instana trace id
        :param span_id: instana span id
        :param level: instana level
        :param synthetic: instana synthetic
        :param traceparent:
        :param tracestate:
        :param disable_w3c_trace_context: flag used to enable w3c trace context only on HTTP requests
        :return: ctx
        """
        correlation = False
        disable_traceparent = os.environ.get("INSTANA_DISABLE_W3C_TRACE_CORRELATION", "")
        instana_ancestor = None
        ctx = SpanContext()
        if level and "correlationType" in level:
            trace_id, span_id = [None] * 2
            correlation = True

        ctx_level = self._get_ctx_level(level)

        if trace_id and span_id:
            ctx.trace_id = trace_id[-16:]  # only the last 16 chars
            ctx.span_id = span_id[-16:]  # only the last 16 chars
            ctx.level = ctx_level
            ctx.synthetic = synthetic is not None

            if len(trace_id) > 16:
                ctx.long_trace_id = trace_id

        elif not disable_w3c_trace_context and traceparent and trace_id is None and span_id is None:
            _, tp_trace_id, tp_parent_id, _ = self._tp.get_traceparent_fields(traceparent)

            if tracestate and "in=" in tracestate:
                instana_ancestor = self._ts.get_instana_ancestor(tracestate)

            if disable_traceparent == "":
                ctx.trace_id = tp_trace_id[-16:]
                ctx.span_id = tp_parent_id
                ctx.level = ctx_level
                ctx.synthetic = synthetic is not None
                ctx.trace_parent = True
                ctx.instana_ancestor = instana_ancestor
                ctx.long_trace_id = tp_trace_id
            else:
                if instana_ancestor:
                    ctx.trace_id = instana_ancestor.t
                    ctx.span_id = instana_ancestor.p
                    ctx.level = ctx_level
                    ctx.synthetic = synthetic is not None

        elif synthetic:
            ctx.synthetic = synthetic

        if correlation:
            self._set_correlation_properties(level, ctx)

        if traceparent:
            ctx.traceparent = traceparent
            ctx.tracestate = tracestate

        return ctx

    def __extract_instana_headers(self, dc):
        """
        Search carrier for the *HEADER* keys and return the tracing key-values

        :param dc: The dict or list potentially containing context
        :return: trace_id, span_id, level, synthetic
        """
        trace_id, span_id, level, synthetic = [None] * 4

        # Headers can exist in the standard X-Instana-T/S format or the alternate HTTP_X_INSTANA_T/S style
        try:
            trace_id = dc.get(self.LC_HEADER_KEY_T) or dc.get(self.ALT_LC_HEADER_KEY_T) or dc.get(
                self.B_HEADER_KEY_T) or dc.get(self.B_ALT_LC_HEADER_KEY_T)
            if trace_id:
                trace_id = header_to_long_id(trace_id)

            span_id = dc.get(self.LC_HEADER_KEY_S) or dc.get(self.ALT_LC_HEADER_KEY_S) or dc.get(
                self.B_HEADER_KEY_S) or dc.get(self.B_ALT_LC_HEADER_KEY_S)
            if span_id:
                span_id = header_to_id(span_id)

            level = dc.get(self.LC_HEADER_KEY_L) or dc.get(self.ALT_LC_HEADER_KEY_L) or dc.get(
                self.B_HEADER_KEY_L) or dc.get(self.B_ALT_LC_HEADER_KEY_L)
            if level and PY3 is True and isinstance(level, bytes):
                level = level.decode("utf-8")

            synthetic = dc.get(self.LC_HEADER_KEY_SYNTHETIC) or dc.get(self.ALT_LC_HEADER_KEY_SYNTHETIC) or dc.get(
                self.B_HEADER_KEY_SYNTHETIC) or dc.get(self.B_ALT_LC_HEADER_KEY_SYNTHETIC)
            if synthetic:
                synthetic = synthetic in ['1', b'1']

        except Exception:
            logger.debug("extract error:", exc_info=True)

        return trace_id, span_id, level, synthetic

    def __extract_w3c_trace_context_headers(self, dc):
        """
        Search carrier for the *HEADER* keys and return the tracing key-values

        :param dc: The dict or list potentially containing context
        :return: traceparent, tracestate
        """
        traceparent, tracestate = [None] * 2

        try:
            traceparent = dc.get(self.HEADER_KEY_TRACEPARENT) or dc.get(self.ALT_HEADER_KEY_TRACEPARENT) or dc.get(
                self.B_HEADER_KEY_TRACEPARENT) or dc.get(self.B_ALT_HEADER_KEY_TRACEPARENT)
            if traceparent and PY3 is True and isinstance(traceparent, bytes):
                traceparent = traceparent.decode("utf-8")

            tracestate = dc.get(self.HEADER_KEY_TRACESTATE) or dc.get(self.ALT_HEADER_KEY_TRACESTATE) or dc.get(
                self.B_HEADER_KEY_TRACESTATE) or dc.get(self.B_ALT_HEADER_KEY_TRACESTATE)
            if tracestate and PY3 is True and isinstance(tracestate, bytes):
                tracestate = tracestate.decode("utf-8")

        except Exception:
            logger.debug("extract error:", exc_info=True)

        return traceparent, tracestate

    def extract(self, carrier, disable_w3c_trace_context=False):
        """
        This method overrides the one of the Baseclass as with the introduction of W3C trace context for the HTTP
        requests more extracting steps and logic was required
        :param disable_w3c_trace_context:
        :param carrier:
        :return: the context or None
        """
        try:
            traceparent, tracestate = [None] * 2
            headers = self._extract_headers_dict(carrier=carrier)
            if headers is None:
                return None
            headers = {k.lower(): v for k, v in headers.items()}

            trace_id, span_id, level, synthetic = self.__extract_instana_headers(dc=headers)
            if not disable_w3c_trace_context:
                traceparent, tracestate = self.__extract_w3c_trace_context_headers(dc=headers)

            if traceparent:
                traceparent = self._tp.validate(traceparent)

            ctx = self.__determine_span_context(trace_id, span_id, level, synthetic, traceparent, tracestate,
                                                disable_w3c_trace_context)

            return ctx
        except Exception:
            logger.debug("extract error:", exc_info=True)


# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import time
import pytest
import requests
import multiprocessing
from instana.singletons import tracer
from ..helpers import testenv
from ..helpers import get_first_span_by_filter


@pytest.fixture(scope="module")
def server():
    from tests.apps.fastapi_app import launch_fastapi
    proc = multiprocessing.Process(target=launch_fastapi, args=(), daemon=True)
    proc.start()
    time.sleep(2)
    yield
    proc.kill()  # Kill server after tests


def test_vanilla_get(server):
    result = requests.get(testenv["fastapi_server"] + '/')

    assert result.status_code == 200
    assert "X-INSTANA-T" in result.headers
    assert "X-INSTANA-S" in result.headers
    assert "X-INSTANA-L" in result.headers
    assert result.headers["X-INSTANA-L"] == '1'
    assert "Server-Timing" in result.headers

    spans = tracer.recorder.queued_spans()
    # FastAPI instrumentation (like all instrumentation) _always_ traces unless told otherwise
    assert len(spans) == 1
    assert spans[0].n == 'asgi'


def test_basic_get(server):
    result = None
    with tracer.start_active_span('test'):
        result = requests.get(testenv["fastapi_server"] + '/')

    assert result.status_code == 200

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 3

    span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
    test_span = get_first_span_by_filter(spans, span_filter)
    assert (test_span)

    span_filter = lambda span: span.n == "urllib3"
    urllib3_span = get_first_span_by_filter(spans, span_filter)
    assert (urllib3_span)

    span_filter = lambda span: span.n == "asgi"
    asgi_span = get_first_span_by_filter(spans, span_filter)
    assert (asgi_span)

    assert (test_span.t == urllib3_span.t == asgi_span.t)
    assert (asgi_span.p == urllib3_span.s)
    assert (urllib3_span.p == test_span.s)

    assert "X-INSTANA-T" in result.headers
    assert result.headers["X-INSTANA-T"] == asgi_span.t
    assert "X-INSTANA-S" in result.headers
    assert result.headers["X-INSTANA-S"] == asgi_span.s
    assert "X-INSTANA-L" in result.headers
    assert result.headers["X-INSTANA-L"] == '1'
    assert "Server-Timing" in result.headers
    assert result.headers["Server-Timing"] == ("intid;desc=%s" % asgi_span.t)

    assert (asgi_span.ec == None)
    assert (asgi_span.data['http']['host'] == '127.0.0.1')
    assert (asgi_span.data['http']['path'] == '/')
    assert (asgi_span.data['http']['path_tpl'] == '/')
    assert (asgi_span.data['http']['method'] == 'GET')
    assert (asgi_span.data['http']['status'] == 200)
    assert (asgi_span.data['http']['error'] is None)
    assert (asgi_span.data['http']['params'] is None)


def test_400(server):
    result = None
    with tracer.start_active_span('test'):
        result = requests.get(testenv["fastapi_server"] + '/400')

    assert result.status_code == 400

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 3

    span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
    test_span = get_first_span_by_filter(spans, span_filter)
    assert (test_span)

    span_filter = lambda span: span.n == "urllib3"
    urllib3_span = get_first_span_by_filter(spans, span_filter)
    assert (urllib3_span)

    span_filter = lambda span: span.n == 'asgi'
    asgi_span = get_first_span_by_filter(spans, span_filter)
    assert (asgi_span)

    assert (test_span.t == urllib3_span.t == asgi_span.t)
    assert (asgi_span.p == urllib3_span.s)
    assert (urllib3_span.p == test_span.s)

    assert "X-INSTANA-T" in result.headers
    assert result.headers["X-INSTANA-T"] == asgi_span.t
    assert "X-INSTANA-S" in result.headers
    assert result.headers["X-INSTANA-S"] == asgi_span.s
    assert "X-INSTANA-L" in result.headers
    assert result.headers["X-INSTANA-L"] == '1'
    assert "Server-Timing" in result.headers
    assert result.headers["Server-Timing"] == ("intid;desc=%s" % asgi_span.t)

    assert (asgi_span.ec == None)
    assert (asgi_span.data['http']['host'] == '127.0.0.1')
    assert (asgi_span.data['http']['path'] == '/400')
    assert (asgi_span.data['http']['path_tpl'] == '/400')
    assert (asgi_span.data['http']['method'] == 'GET')
    assert (asgi_span.data['http']['status'] == 400)
    assert (asgi_span.data['http']['error'] is None)
    assert (asgi_span.data['http']['params'] is None)

def test_500(server):
    result = None
    with tracer.start_active_span('test'):
        result = requests.get(testenv["fastapi_server"] + '/500')

    assert result.status_code == 500

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 3

    span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
    test_span = get_first_span_by_filter(spans, span_filter)
    assert (test_span)

    span_filter = lambda span: span.n == "urllib3"
    urllib3_span = get_first_span_by_filter(spans, span_filter)
    assert (urllib3_span)

    span_filter = lambda span: span.n == 'asgi'
    asgi_span = get_first_span_by_filter(spans, span_filter)
    assert (asgi_span)

    assert (test_span.t == urllib3_span.t == asgi_span.t)
    assert (asgi_span.p == urllib3_span.s)
    assert (urllib3_span.p == test_span.s)

    assert "X-INSTANA-T" in result.headers
    assert result.headers["X-INSTANA-T"] == asgi_span.t
    assert "X-INSTANA-S" in result.headers
    assert result.headers["X-INSTANA-S"] == asgi_span.s
    assert "X-INSTANA-L" in result.headers
    assert result.headers["X-INSTANA-L"] == '1'
    assert "Server-Timing" in result.headers
    assert result.headers["Server-Timing"] == ("intid;desc=%s" % asgi_span.t)

    assert (asgi_span.ec == 1)
    assert (asgi_span.data['http']['host'] == '127.0.0.1')
    assert (asgi_span.data['http']['path'] == '/500')
    assert (asgi_span.data['http']['path_tpl'] == '/500')
    assert (asgi_span.data['http']['method'] == 'GET')
    assert (asgi_span.data['http']['status'] == 500)
    assert (asgi_span.data['http']['error'] == '500 response')
    assert (asgi_span.data['http']['params'] is None)

def test_path_templates(server):
    result = None
    with tracer.start_active_span('test'):
        result = requests.get(testenv["fastapi_server"] + '/users/1')

    assert result.status_code == 200

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 3

    span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
    test_span = get_first_span_by_filter(spans, span_filter)
    assert (test_span)

    span_filter = lambda span: span.n == "urllib3"
    urllib3_span = get_first_span_by_filter(spans, span_filter)
    assert (urllib3_span)

    span_filter = lambda span: span.n == 'asgi'
    asgi_span = get_first_span_by_filter(spans, span_filter)
    assert (asgi_span)

    assert (test_span.t == urllib3_span.t == asgi_span.t)
    assert (asgi_span.p == urllib3_span.s)
    assert (urllib3_span.p == test_span.s)

    assert "X-INSTANA-T" in result.headers
    assert result.headers["X-INSTANA-T"] == asgi_span.t
    assert "X-INSTANA-S" in result.headers
    assert result.headers["X-INSTANA-S"] == asgi_span.s
    assert "X-INSTANA-L" in result.headers
    assert result.headers["X-INSTANA-L"] == '1'
    assert "Server-Timing" in result.headers
    assert result.headers["Server-Timing"] == ("intid;desc=%s" % asgi_span.t)

    assert (asgi_span.ec == None)
    assert (asgi_span.data['http']['host'] == '127.0.0.1')
    assert (asgi_span.data['http']['path'] == '/users/1')
    assert (asgi_span.data['http']['path_tpl'] == '/users/{user_id}')
    assert (asgi_span.data['http']['method'] == 'GET')
    assert (asgi_span.data['http']['status'] == 200)
    assert (asgi_span.data['http']['error'] is None)
    assert (asgi_span.data['http']['params'] is None)


def test_secret_scrubbing(server):
    result = None
    with tracer.start_active_span('test'):
        result = requests.get(testenv["fastapi_server"] + '/?secret=shhh')

    assert result.status_code == 200

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 3

    span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
    test_span = get_first_span_by_filter(spans, span_filter)
    assert (test_span)

    span_filter = lambda span: span.n == "urllib3"
    urllib3_span = get_first_span_by_filter(spans, span_filter)
    assert (urllib3_span)

    span_filter = lambda span: span.n == 'asgi'
    asgi_span = get_first_span_by_filter(spans, span_filter)
    assert (asgi_span)

    assert (test_span.t == urllib3_span.t == asgi_span.t)
    assert (asgi_span.p == urllib3_span.s)
    assert (urllib3_span.p == test_span.s)

    assert "X-INSTANA-T" in result.headers
    assert result.headers["X-INSTANA-T"] == asgi_span.t
    assert "X-INSTANA-S" in result.headers
    assert result.headers["X-INSTANA-S"] == asgi_span.s
    assert "X-INSTANA-L" in result.headers
    assert result.headers["X-INSTANA-L"] == '1'
    assert "Server-Timing" in result.headers
    assert result.headers["Server-Timing"] == ("intid;desc=%s" % asgi_span.t)

    assert (asgi_span.ec == None)
    assert (asgi_span.data['http']['host'] == '127.0.0.1')
    assert (asgi_span.data['http']['path'] == '/')
    assert (asgi_span.data['http']['path_tpl'] == '/')
    assert (asgi_span.data['http']['method'] == 'GET')
    assert (asgi_span.data['http']['status'] == 200)
    assert (asgi_span.data['http']['error'] is None)
    assert (asgi_span.data['http']['params'] == 'secret=<redacted>')

def test_synthetic_request(server):
    request_headers = {
        'X-INSTANA-SYNTHETIC': '1'
    }
    with tracer.start_active_span('test'):
        result = requests.get(testenv["fastapi_server"] + '/', headers=request_headers)

    assert result.status_code == 200

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 3

    span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
    test_span = get_first_span_by_filter(spans, span_filter)
    assert (test_span)

    span_filter = lambda span: span.n == "urllib3"
    urllib3_span = get_first_span_by_filter(spans, span_filter)
    assert (urllib3_span)

    span_filter = lambda span: span.n == 'asgi'
    asgi_span = get_first_span_by_filter(spans, span_filter)
    assert (asgi_span)

    assert (test_span.t == urllib3_span.t == asgi_span.t)
    assert (asgi_span.p == urllib3_span.s)
    assert (urllib3_span.p == test_span.s)

    assert "X-INSTANA-T" in result.headers
    assert result.headers["X-INSTANA-T"] == asgi_span.t
    assert "X-INSTANA-S" in result.headers
    assert result.headers["X-INSTANA-S"] == asgi_span.s
    assert "X-INSTANA-L" in result.headers
    assert result.headers["X-INSTANA-L"] == '1'
    assert "Server-Timing" in result.headers
    assert result.headers["Server-Timing"] == ("intid;desc=%s" % asgi_span.t)

    assert (asgi_span.ec == None)
    assert (asgi_span.data['http']['host'] == '127.0.0.1')
    assert (asgi_span.data['http']['path'] == '/')
    assert (asgi_span.data['http']['path_tpl'] == '/')
    assert (asgi_span.data['http']['method'] == 'GET')
    assert (asgi_span.data['http']['status'] == 200)
    assert (asgi_span.data['http']['error'] is None)
    assert (asgi_span.data['http']['params'] is None)

    assert (asgi_span.sy)
    assert (urllib3_span.sy is None)
    assert (test_span.sy is None)


def test_custom_header_capture(server):
    from instana.singletons import agent

    # The background FastAPI server is pre-configured with custom headers to capture

    request_headers = {
        'X-Capture-This': 'this',
        'X-Capture-That': 'that'
    }
    with tracer.start_active_span('test'):
        result = requests.get(testenv["fastapi_server"] + '/', headers=request_headers)

    assert result.status_code == 200

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 3

    span_filter = lambda span: span.n == "sdk" and span.data['sdk']['name'] == 'test'
    test_span = get_first_span_by_filter(spans, span_filter)
    assert (test_span)

    span_filter = lambda span: span.n == "urllib3"
    urllib3_span = get_first_span_by_filter(spans, span_filter)
    assert (urllib3_span)

    span_filter = lambda span: span.n == 'asgi'
    asgi_span = get_first_span_by_filter(spans, span_filter)
    assert (asgi_span)

    assert (test_span.t == urllib3_span.t == asgi_span.t)
    assert (asgi_span.p == urllib3_span.s)
    assert (urllib3_span.p == test_span.s)

    assert "X-INSTANA-T" in result.headers
    assert result.headers["X-INSTANA-T"] == asgi_span.t
    assert "X-INSTANA-S" in result.headers
    assert result.headers["X-INSTANA-S"] == asgi_span.s
    assert "X-INSTANA-L" in result.headers
    assert result.headers["X-INSTANA-L"] == '1'
    assert "Server-Timing" in result.headers
    assert result.headers["Server-Timing"] == ("intid;desc=%s" % asgi_span.t)

    assert (asgi_span.ec == None)
    assert (asgi_span.data['http']['host'] == '127.0.0.1')
    assert (asgi_span.data['http']['path'] == '/')
    assert (asgi_span.data['http']['path_tpl'] == '/')
    assert (asgi_span.data['http']['method'] == 'GET')
    assert (asgi_span.data['http']['status'] == 200)
    assert (asgi_span.data['http']['error'] is None)
    assert (asgi_span.data['http']['params'] is None)

    assert ("X-Capture-This" in asgi_span.data["http"]["header"])
    assert ("this" == asgi_span.data["http"]["header"]["X-Capture-This"])
    assert ("X-Capture-That" in asgi_span.data["http"]["header"])
    assert ("that" == asgi_span.data["http"]["header"]["X-Capture-That"])



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import threading

from ...log import logger
from ..runtime import min_version, runtime_info
from ..profile import Profile
from ..profile import CallSite
from ..schedule import schedule, delay

if min_version(3, 4):
    import tracemalloc


class AllocationSampler(object):
    MAX_TRACEBACK_SIZE = 25 # number of frames
    MAX_MEMORY_OVERHEAD = 10 * 1e6 # 10MB
    MAX_PROFILED_ALLOCATIONS = 25

    def __init__(self, profiler):
        self.profiler = profiler
        self.ready = False
        self.top = None
        self.top_lock = threading.Lock()
        self.overhead_monitor = None

    def setup(self):
        if self.profiler.get_option('allocation_sampler_disabled'):
            return

        if not runtime_info.OS_LINUX and not runtime_info.OS_DARWIN:
            logger.debug('Allocation sampler is only supported on Linux and OS X.')
            return

        if not min_version(3, 4):
            logger.debug('Memory allocation profiling is available for Python 3.4 or higher')
            return

        self.ready = True

    def reset(self):
        self.top = CallSite('', '', 0)

    def start_sampler(self):
        logger.debug('Activating memory allocation sampler.')

        def start():
            tracemalloc.start(self.MAX_TRACEBACK_SIZE)
        self.profiler.run_in_main_thread(start)

        def monitor_overhead():
            if tracemalloc.is_tracing() and tracemalloc.get_tracemalloc_memory() > self.MAX_MEMORY_OVERHEAD:
                logger.debug('Allocation sampler memory overhead limit exceeded: %s bytes', tracemalloc.get_tracemalloc_memory())
                self.stop_sampler()

        if not self.profiler.get_option('disable_timers'):
            self.overhead_monitor = schedule(0.5, 0.5, monitor_overhead)

    def stop_sampler(self):
        logger.debug('Deactivating memory allocation sampler.')

        with self.top_lock:
            if self.overhead_monitor:
                self.overhead_monitor.cancel()
                self.overhead_monitor = None

            if tracemalloc.is_tracing():
                snapshot = tracemalloc.take_snapshot()
                logger.debug('Allocation sampler memory overhead %s bytes', tracemalloc.get_tracemalloc_memory())
                tracemalloc.stop()
                self.process_snapshot(snapshot)

    def build_profile(self, duration, timespan):
        with self.top_lock:
            self.top.normalize(duration)
            self.top.floor()

            profile = Profile(
                Profile.CATEGORY_MEMORY,
                Profile.TYPE_MEMORY_ALLOCATION_RATE,
                Profile.UNIT_BYTE,
                self.top.children.values(),
                duration,
                timespan
            )

            return profile

    def destroy(self):
        pass

    def process_snapshot(self, snapshot):
        stats = snapshot.statistics('traceback')

        for stat in stats[:self.MAX_PROFILED_ALLOCATIONS]:
            if stat.traceback:
                skip_stack = False
                for frame in stat.traceback:
                    if frame.filename and self.profiler.frame_cache.is_profiler_frame(frame.filename):
                        skip_stack = True
                        break
                if skip_stack:
                    continue

                current_node = self.top
                for frame in reversed(stat.traceback):
                    if frame.filename == '<unknown>':
                        continue

                    current_node = current_node.find_or_add_child('', frame.filename, frame.lineno)
                current_node.increment(stat.size, stat.count)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Instrumentation for AWS Lambda functions
"""
import sys
import wrapt

from ...log import logger
from ...singletons import env_is_aws_lambda
from ... import get_lambda_handler_or_default
from ...singletons import get_agent, get_tracer
from .triggers import enrich_lambda_span, get_context
import traceback


def lambda_handler_with_instana(wrapped, instance, args, kwargs):
    event = args[0]
    agent = get_agent()
    tracer = get_tracer()

    agent.collector.collect_snapshot(*args)
    incoming_ctx = get_context(tracer, event)

    result = None
    with tracer.start_active_span("aws.lambda.entry", child_of=incoming_ctx) as scope:
        enrich_lambda_span(agent, scope.span, *args)
        try:
            result = wrapped(*args, **kwargs)

            if isinstance(result, dict):
                server_timing_value = "intid;desc=%s" % scope.span.context.trace_id
                if 'headers' in result:
                    result['headers']['Server-Timing'] = server_timing_value
                elif 'multiValueHeaders' in result:
                    result['multiValueHeaders']['Server-Timing'] = [server_timing_value]
        except Exception as exc:
            if scope.span:
                exc = traceback.format_exc()
                scope.span.log_exception(exc)
            raise
        finally:
            scope.close()
            agent.collector.shutdown()

    agent.collector.shutdown()
    return result


if env_is_aws_lambda is True:
    handler_module, handler_function = get_lambda_handler_or_default()

    if handler_module is not None and handler_function is not None:
        try:
            logger.debug("Instrumenting AWS Lambda handler (%s.%s)" % (handler_module, handler_function))
            sys.path.insert(0, '/var/runtime')
            sys.path.insert(0, '/var/task')
            wrapt.wrap_function_wrapper(handler_module, handler_function, lambda_handler_with_instana)
        except (ModuleNotFoundError, ImportError) as exc:
            logger.warning("Instana: Couldn't instrument AWS Lambda handler. Not monitoring.")
    else:
        logger.warning("Instana: Couldn't determine AWS Lambda Handler.  Not monitoring.")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from ..log import logger

def normalize_aws_lambda_arn(context):
    """
    Parse the AWS Lambda context object for a fully qualified AWS Lambda function ARN.

    This method will ensure that the returned value matches the following ARN pattern:
      arn:aws:lambda:${region}:${account-id}:function:${name}:${version}

    @param context:  AWS Lambda context object
    @return:
    """
    try:
        arn = context.invoked_function_arn
        parts = arn.split(':')

        count = len(parts)
        if count == 7:
            # need to append version
            arn = arn + ':' + context.function_version
        elif count != 8:
            logger.debug("Unexpected ARN parse issue: %s", arn)

        return arn
    except Exception:
        logger.debug("normalize_arn: ", exc_info=True)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from __future__ import absolute_import

import json
import wrapt
from opentracing import Format

from ....log import logger
from ....singletons import tracer

try:
    from google.cloud import pubsub_v1


    def _set_publisher_tags(span, topic_path):
        span.set_tag('gcps.op', 'publish')
        # Fully qualified identifier is in the form of
        # `projects/{project_id}/topic/{topic_name}`
        project_id, topic_name = topic_path.split('/')[1::2]
        span.set_tag('gcps.projid', project_id)
        span.set_tag('gcps.top', topic_name)


    def _set_consumer_tags(span, subscription_path):
        span.set_tag('gcps.op', 'consume')
        # Fully qualified identifier is in the form of
        # `projects/{project_id}/subscriptions/{subscription_name}`
        project_id, subscription_id = subscription_path.split('/')[1::2]
        span.set_tag('gcps.projid', project_id)
        span.set_tag('gcps.sub', subscription_id)


    @wrapt.patch_function_wrapper('google.cloud.pubsub_v1', 'PublisherClient.publish')
    def publish_with_instana(wrapped, instance, args, kwargs):
        """References:
        - PublisherClient.publish(topic_path, messages, metadata)
        """
        # check if active
        parent_span = tracer.active_span

        # return early if we're not tracing
        if parent_span is None:
            return wrapped(*args, **kwargs)

        with tracer.start_active_span('gcps-producer', child_of=parent_span) as scope:
            # trace continuity, inject to the span context
            headers = dict()
            tracer.inject(scope.span.context, Format.TEXT_MAP, headers, disable_w3c_trace_context=True)

            # update the metadata dict with instana trace attributes
            kwargs.update(headers)

            _set_publisher_tags(scope.span, topic_path=args[0])

            try:
                rv = wrapped(*args, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return rv


    @wrapt.patch_function_wrapper('google.cloud.pubsub_v1', 'SubscriberClient.subscribe')
    def subscribe_with_instana(wrapped, instance, args, kwargs):

        """References:
        - SubscriberClient.subscribe(subscription_path, callback)
        - callback(message) is called from the subscription future
        """

        def callback_with_instana(message):
            if message.attributes:
                parent_span = tracer.extract(Format.TEXT_MAP, message.attributes, disable_w3c_trace_context=True)
            else:
                parent_span = None

            with tracer.start_active_span('gcps-consumer', child_of=parent_span) as scope:
                _set_consumer_tags(scope.span, subscription_path=args[0])
                try:
                    callback(message)
                except Exception as e:
                    scope.span.log_exception(e)
                    raise

        # Handle callback appropriately from args or kwargs
        if 'callback' in kwargs:
            callback = kwargs.get('callback')
            kwargs['callback'] = callback_with_instana
            return wrapped(*args, **kwargs)
        else:
            subscription, callback, *args = args
            args = (subscription, callback_with_instana, *args)
            return wrapped(*args, **kwargs)


    logger.debug('Instrumenting Google Cloud Pub/Sub')
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Base class for the various helpers that can be used by Collectors.  Helpers assist
in the data collection for various entities such as host, hardware, AWS Task, ec2,
memory, cpu, docker etc etc..
"""
from ...log import logger


class BaseHelper(object):
    """
    Base class for all helpers.  Descendants must override and implement `self.collect_metrics`.
    """
    def __init__(self, collector):
        self.collector = collector

    def get_delta(self, source, previous, metric):
        """
        Given a metric, see if the value varies from the previous reported metrics

        @param source [dict or value]: the dict to retrieve the new value of <metric> (as source[metric]) or
          if not a dict, then the new value of the metric
        @param previous [dict]: the previous value of <metric> that was reported (as previous[metric])
        @param metric [String or Tuple]: the name of the metric in question.  If the keys for source[metric],
          and previous[metric] vary, you can pass a tuple in the form of (src, dst)
        @return: None (meaning no difference) or the new value (source[metric])
        """
        if isinstance(metric, tuple):
            src_metric = metric[0]
            dst_metric = metric[1]
        else:
            src_metric = metric
            dst_metric = metric

        if isinstance(source, dict):
            new_value = source.get(src_metric, None)
        else:
            new_value = source

        if previous[dst_metric] != new_value:
            return new_value
        else:
            return None

    def apply_delta(self, source, previous, new, metric, with_snapshot):
        """
        Helper method to assist in delta reporting of metrics.

        @param source [dict or value]: the dict to retrieve the new value of <metric> (as source[metric]) or
          if not a dict, then the new value of the metric
        @param previous [dict]: the previous value of <metric> that was reported (as previous[metric])
        @param new [dict]: the new value of the metric that will be sent new (as new[metric])
        @param metric [String or Tuple]: the name of the metric in question.  If the keys for source[metric],
          previous[metric] and new[metric] vary, you can pass a tuple in the form of (src, dst)
        @param with_snapshot [Bool]: if this metric is being sent with snapshot data
        @return: None
        """
        if isinstance(metric, tuple):
            src_metric = metric[0]
            dst_metric = metric[1]
        else:
            src_metric = metric
            dst_metric = metric

        if isinstance(source, dict):
            new_value = source.get(src_metric, None)
        else:
            new_value = source

        previous_value = previous.get(dst_metric, 0)

        if previous_value != new_value or with_snapshot is True:
            previous[dst_metric] = new[dst_metric] = new_value
    
    def collect_metrics(self, **kwargs):
        logger.debug("BaseHelper.collect_metrics must be overridden")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2018

from __future__ import absolute_import

import re
from operator import attrgetter

from ..log import logger
from ..util.traceutils import get_active_tracer

try:
    import sqlalchemy
    from sqlalchemy import event
    from sqlalchemy.engine import Engine

    url_regexp = re.compile(r"\/\/(\S+@)")


    @event.listens_for(Engine, 'before_cursor_execute', named=True)
    def receive_before_cursor_execute(**kw):
        try:
            active_tracer = get_active_tracer()

            # If we're not tracing, just return
            if active_tracer is None:
                return

            scope = active_tracer.start_active_span("sqlalchemy", child_of=active_tracer.active_span)
            context = kw['context']
            if context:
                context._stan_scope = scope

            conn = kw['conn']
            url = str(conn.engine.url)
            scope.span.set_tag('sqlalchemy.sql', kw['statement'])
            scope.span.set_tag('sqlalchemy.eng', conn.engine.name)
            scope.span.set_tag('sqlalchemy.url', url_regexp.sub('//', url))
        except Exception as e:
            logger.debug(e)
        return


    @event.listens_for(Engine, 'after_cursor_execute', named=True)
    def receive_after_cursor_execute(**kw):
        context = kw['context']

        if context is not None and hasattr(context, '_stan_scope'):
            scope = context._stan_scope
            if scope is not None:
                scope.close()


    error_event = "handle_error"
    # Handle dbapi_error event; deprecated since version 0.9
    if sqlalchemy.__version__[0] == "0":
        error_event = "dbapi_error"


    def _set_error_tags(context, exception_string, scope_string):
        scope, context_exception = None, None
        if attrgetter(scope_string)(context) and attrgetter(exception_string)(context):
            scope = attrgetter(scope_string)(context)
            context_exception = attrgetter(exception_string)(context)
        if scope and context_exception:
            scope.span.log_exception(context_exception)
            scope.close()
        else:
            scope.span.log_exception("No %s specified." % error_event)
            scope.close()


    @event.listens_for(Engine, error_event, named=True)
    def receive_handle_db_error(**kw):

        # support older db error event
        if error_event == "dbapi_error":
            context = kw.get('context')
            exception_string = 'exception'
            scope_string = '_stan_scope'
        else:
            context = kw.get('exception_context')
            exception_string = 'sqlalchemy_exception'
            scope_string = 'execution_context._stan_scope'

        if context:
            _set_error_tags(context, exception_string, scope_string)


    logger.debug("Instrumenting sqlalchemy")

except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from __future__ import absolute_import

import os
import json
import requests_mock
import unittest

from instana.tracer import InstanaTracer
from instana.recorder import StanRecorder
from instana.agent.google_cloud_run import GCRAgent
from instana.singletons import get_agent, set_agent, get_tracer, set_tracer


class TestGCRCollector(unittest.TestCase):
    def __init__(self, methodName='runTest'):
        super(TestGCRCollector, self).__init__(methodName)
        self.agent = None
        self.span_recorder = None
        self.tracer = None
        self.pwd = os.path.dirname(os.path.realpath(__file__))

        self.original_agent = get_agent()
        self.original_tracer = get_tracer()

    def setUp(self):
        os.environ["PORT"] = "port"
        os.environ["INSTANA_ENDPOINT_URL"] = "https://localhost/notreal"
        os.environ["INSTANA_AGENT_KEY"] = "Fake_Key"

        if "INSTANA_ZONE" in os.environ:
            os.environ.pop("INSTANA_ZONE")
        if "INSTANA_TAGS" in os.environ:
            os.environ.pop("INSTANA_TAGS")

    def tearDown(self):
        """ Reset all environment variables of consequence """
        if "PORT" in os.environ:
            os.environ.pop("PORT")
        if "INSTANA_EXTRA_HTTP_HEADERS" in os.environ:
            os.environ.pop("INSTANA_EXTRA_HTTP_HEADERS")
        if "INSTANA_ENDPOINT_URL" in os.environ:
            os.environ.pop("INSTANA_ENDPOINT_URL")
        if "INSTANA_AGENT_KEY" in os.environ:
            os.environ.pop("INSTANA_AGENT_KEY")
        if "INSTANA_ZONE" in os.environ:
            os.environ.pop("INSTANA_ZONE")
        if "INSTANA_TAGS" in os.environ:
            os.environ.pop("INSTANA_TAGS")

        set_agent(self.original_agent)
        set_tracer(self.original_tracer)

    def create_agent_and_setup_tracer(self):
        self.agent = GCRAgent(service="service", configuration="configuration", revision="revision")
        self.span_recorder = StanRecorder(self.agent)
        self.tracer = InstanaTracer(recorder=self.span_recorder)
        set_agent(self.agent)
        set_tracer(self.tracer)

        # Manually set the Instance and Project Metadata API results on the collector
        with open(self.pwd + '/../data/gcr/instance_metadata.json', 'r') as json_file:
            self.agent.collector.instance_metadata = json.load(json_file)
        with open(self.pwd + '/../data/gcr/project_metadata.json', 'r') as json_file:
            self.agent.collector.project_metadata = json.load(json_file)

    @requests_mock.Mocker()
    def test_prepare_payload_basics(self, m):
        self.create_agent_and_setup_tracer()
        m.get("http://metadata.google.internal/computeMetadata/v1/project/?recursive=true",
              headers={"Metadata-Flavor": "Google"}, json=self.agent.collector.project_metadata)

        m.get("http://metadata.google.internal/computeMetadata/v1/instance/?recursive=true",
              headers={"Metadata-Flavor": "Google"}, json=self.agent.collector.instance_metadata)

        payload = self.agent.collector.prepare_payload()
        assert (payload)

        assert (len(payload.keys()) == 2)
        assert ('spans' in payload)
        assert (isinstance(payload['spans'], list))
        assert (len(payload['spans']) == 0)
        assert ('metrics' in payload)
        assert (len(payload['metrics'].keys()) == 1)
        assert ('plugins' in payload['metrics'])
        assert (isinstance(payload['metrics']['plugins'], list))
        assert (len(payload['metrics']['plugins']) == 2)

        plugins = payload['metrics']['plugins']
        for plugin in plugins:
            # print("%s - %s" % (plugin["name"], plugin["entityId"]))
            assert ('name' in plugin)
            assert ('entityId' in plugin)
            assert ('data' in plugin)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

from __future__ import absolute_import

import wrapt
import flask
import opentracing
import opentracing.ext.tags as ext

from ...log import logger
from ...singletons import tracer


@wrapt.patch_function_wrapper('flask', 'templating._render')
def render_with_instana(wrapped, instance, argv, kwargs):
    ctx = argv[1]

    # If we're not tracing, just return
    if not hasattr(ctx['g'], 'scope'):
        return wrapped(*argv, **kwargs)

    with tracer.start_active_span("render", child_of=ctx['g'].scope.span) as rscope:
        try:
            template = argv[0]

            rscope.span.set_tag("type", "template")
            if template.name is None:
                rscope.span.set_tag("name", '(from string)')
            else:
                rscope.span.set_tag("name", template.name)

            return wrapped(*argv, **kwargs)
        except Exception as e:
            rscope.span.log_exception(e)
            raise


@wrapt.patch_function_wrapper('flask', 'Flask.handle_user_exception')
def handle_user_exception_with_instana(wrapped, instance, argv, kwargs):
    # Call original and then try to do post processing
    response = wrapped(*argv, **kwargs)

    try:
        exc = argv[0]

        if hasattr(flask.g, 'scope') and flask.g.scope is not None:
            scope = flask.g.scope
            span = scope.span

            if response is not None:
                if isinstance(response, tuple):
                    status_code = response[1]
                else:
                    if hasattr(response, 'code'):
                        status_code = response.code
                    else:
                        status_code = response.status_code

                if 500 <= status_code <= 511:
                    span.log_exception(exc)

                span.set_tag(ext.HTTP_STATUS_CODE, int(status_code))

                if hasattr(response, 'headers'):
                    tracer.inject(scope.span.context, opentracing.Format.HTTP_HEADERS, response.headers)
                    value = "intid;desc=%s" % scope.span.context.trace_id
                    if hasattr(response.headers, 'add'):
                        response.headers.add('Server-Timing', value)
                    elif type(response.headers) is dict or hasattr(response.headers, "__dict__"):
                        response.headers['Server-Timing'] = value

            scope.close()
            flask.g.scope = None
    except:
        logger.debug("handle_user_exception_with_instana:", exc_info=True)

    return response



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from wsgiref.simple_server import make_server
from pyramid.config import Configurator
import logging

from pyramid.response import Response
import pyramid.httpexceptions as exc

from ...helpers import testenv

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

testenv["pyramid_port"] = 10815
testenv["pyramid_server"] = ("http://127.0.0.1:" + str(testenv["pyramid_port"]))

def hello_world(request):
    return Response('Ok')

def please_fail(request):
    raise exc.HTTPInternalServerError("internal error")

def tableflip(request):
    raise BaseException("fake exception")

app = None
with Configurator() as config:
    config.add_tween('instana.instrumentation.pyramid.tweens.InstanaTweenFactory')
    config.add_route('hello', '/')
    config.add_view(hello_world, route_name='hello')
    config.add_route('fail', '/500')
    config.add_view(please_fail, route_name='fail')
    config.add_route('crash', '/exception')
    config.add_view(tableflip, route_name='crash')
    app = config.make_wsgi_app()
    
pyramid_server = make_server('127.0.0.1', testenv["pyramid_port"], app)




# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import threading
import os
import signal
import atexit
import platform

from ..log import logger
from .runtime import min_version, runtime_info, register_signal
from .frame_cache import FrameCache
from .sampler_scheduler import SamplerScheduler, SamplerConfig
from .samplers.cpu_sampler import CPUSampler
from .samplers.allocation_sampler import AllocationSampler
from .samplers.block_sampler import BlockSampler


class Profiler(object):

    def __init__(self, agent):
        self.agent = agent

        self.profiler_started = False
        self.profiler_destroyed = False

        self.sampler_active = False

        self.main_thread_func = None

        self.frame_cache = FrameCache(self)

        config = SamplerConfig()
        config.log_prefix = 'CPU sampler'
        config.max_profile_duration = 20
        config.max_span_duration = 5
        config.max_span_count = 30
        config.span_interval = 20
        config.report_interval = 120
        self.cpu_sampler_scheduler = SamplerScheduler(self, CPUSampler(self), config)

        config = SamplerConfig()
        config.log_prefix = 'Allocation sampler'
        config.max_profile_duration = 20
        config.max_span_duration = 5
        config.max_span_count = 30
        config.span_interval = 20
        config.report_interval = 120
        self.allocation_sampler_scheduler = SamplerScheduler(self, AllocationSampler(self), config)

        config = SamplerConfig()
        config.log_prefix = 'Block sampler'
        config.max_profile_duration = 20
        config.max_span_duration = 5
        config.max_span_count = 30
        config.span_interval = 20
        config.report_interval = 120
        self.block_sampler_scheduler = SamplerScheduler(self, BlockSampler(self), config)

        self.options = None

    def get_option(self, name, default_val=None):
        if name not in self.options:
            return default_val
        else:
            return self.options[name]

    def start(self, **kwargs):
        if self.profiler_started:
            return

        try:
            if not min_version(2, 7) and not min_version(3, 4):
                raise Exception('Supported Python versions 2.6 or higher and 3.4 or higher')

            if platform.python_implementation() != 'CPython':
                raise Exception('Supported Python interpreter is CPython')

            if self.profiler_destroyed:
                logger.warning('Destroyed profiler cannot be started')
                return

            self.options = kwargs

            self.frame_cache.start()

            self.cpu_sampler_scheduler.setup()
            self.allocation_sampler_scheduler.setup()
            self.block_sampler_scheduler.setup()

            # execute main_thread_func in main thread on signal
            def _signal_handler(signum, frame):
                if(self.main_thread_func):
                    func = self.main_thread_func
                    self.main_thread_func = None
                    try:
                        func()
                    except Exception:
                        logger.error('Error in signal handler function', exc_info=True)

                    return True

            if not runtime_info.OS_WIN:
                register_signal(signal.SIGUSR2, _signal_handler)

            self.cpu_sampler_scheduler.start()
            self.allocation_sampler_scheduler.start()
            self.block_sampler_scheduler.start()

            self.profiler_started = True
            logger.debug('Profiler started')
        except Exception:
            logger.error('Error starting profiler', exc_info=True)

    def destroy(self):
        if not self.profiler_started:
            logger.warning('Profiler has not been started')
            return

        if self.profiler_destroyed:
            return

        self.frame_cache.stop()
        self.cpu_sampler_scheduler.stop()
        self.allocation_sampler_scheduler.stop()
        self.block_sampler_scheduler.stop()

        self.cpu_sampler_scheduler.destroy()
        self.allocation_sampler_scheduler.destroy()
        self.block_sampler_scheduler.destroy()

        self.profiler_destroyed = True
        logger.debug('Profiler destroyed')

    def run_in_thread(self, func):
        def func_wrapper():
            try:
                func()
            except Exception:
                logger.error('Error in thread function', exc_info=True)

        t = threading.Thread(target=func_wrapper)
        t.start()
        return t

    def run_in_main_thread(self, func):
        if self.main_thread_func:
            return False

        self.main_thread_func = func
        os.kill(os.getpid(), signal.SIGUSR2)

        return True



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

"""
This module provides "python -m instana" functionality.  This is used for basic module
information display and a IPython console to diagnose environments.

The console is disabled by default unless the ipython package is installed.
"""
import os
import sys

print("""\
============================================================================
8888888 888b    888  .d8888b. 88888888888     d8888 888b    888        d8888
  888   8888b   888 d88P  Y88b    888        d88888 8888b   888       d88888
  888   88888b  888 Y88b.         888       d88P888 88888b  888      d88P888
  888   888Y88b 888  "Y888b.      888      d88P 888 888Y88b 888     d88P 888
  888   888 Y88b888     "Y88b.    888     d88P  888 888 Y88b888    d88P  888
  888   888  Y88888       "888    888    d88P   888 888  Y88888   d88P   888
  888   888   Y8888 Y88b  d88P    888   d8888888888 888   Y8888  d8888888888
8888888 888    Y888  "Y8888P"     888  d88P     888 888    Y888 d88P     888
============================================================================
""")

if "console" in sys.argv:
    try:
        import IPython
    except ImportError:
        print("This console is not enabled by default.")
        print("IPython not installed.  To use this debug console do: 'pip install ipython'\n")
    else:
        print("Welcome to the Instana console.\n")
        print("This is a simple IPython console with the Instana Python Sensor pre-loaded.\n")

        if "INSTANA_DEBUG" not in os.environ:
            print("If you want debug output of this sensors' activity run instead:\n")
            print("   INSTANA_DEBUG=true python -m instana console")

        print("""
Helpful Links
============================================================================

Monitoring Python Documentation:
https://www.instana.com/docs/ecosystem/python/


Help & Support:
https://support.instana.com/
""")

        IPython.start_ipython(argv=[])
else:
    print("""\
This is an informational screen for Instana.

Supported commands:
 - console: 
   * Requires ipython package: pip install ipython
   * Example: 
     - python -m instana console

See the Instana Python documentation for details on using this package with
your Python applications, workers, queues, neural networks and more.


Related Blog Posts:
============================================================================

Monitoring Python with Instana
https://www.instana.com/blog/monitoring-python-instana/

Zero-Effort, Fully Automatic Distributed Tracing for Python
https://www.instana.com/blog/zero-effort-fully-automatic-distributed-tracing-for-python/


Helpful Links
============================================================================

Monitoring Python Documentation:
https://www.instana.com/docs/ecosystem/python/

Help & Support:
https://support.instana.com/

Python Instrumentation on Github:
https://github.com/instana/python-sensor/
""")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2018

import os
import sys
import time
import opentracing
import opentracing.ext.tags as ext

from django.conf.urls import url
from django.http import HttpResponse, Http404

filepath, extension = os.path.splitext(__file__)
os.environ['DJANGO_SETTINGS_MODULE'] = os.path.basename(filepath)
sys.path.insert(0, os.path.dirname(os.path.abspath(filepath)))

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SECRET_KEY = '^(myu#*^5v-9o$i-%6vnlwvy^#7&hspj$m3lcq#b$@__@+zd@c'
DEBUG = True
ALLOWED_HOSTS = ['testserver', 'localhost']
INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
]
MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]
ROOT_URLCONF = 'app_django'
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]
WSGI_APPLICATION = 'app_django.wsgi.application'
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}
AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]
LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True
STATIC_URL = '/static/'


def index(request):
    return HttpResponse('Stan wuz here!')


def cause_error(request):
    raise Exception('This is a fake error: /cause-error')


def another(request):
    return HttpResponse('Stan wuz here!')


def not_found(request):
    raise Http404('Nothing here')


def complex(request):
    with opentracing.tracer.start_active_span('asteroid') as pscope:
        pscope.span.set_tag(ext.COMPONENT, "Python simple example app")
        pscope.span.set_tag(ext.SPAN_KIND, ext.SPAN_KIND_RPC_SERVER)
        pscope.span.set_tag(ext.PEER_HOSTNAME, "localhost")
        pscope.span.set_tag(ext.HTTP_URL, "/python/simple/one")
        pscope.span.set_tag(ext.HTTP_METHOD, "GET")
        pscope.span.set_tag(ext.HTTP_STATUS_CODE, 200)
        pscope.span.log_kv({"foo": "bar"})
        time.sleep(.2)

        with opentracing.tracer.start_active_span('spacedust', child_of=pscope.span) as cscope:
            cscope.span.set_tag(ext.SPAN_KIND, ext.SPAN_KIND_RPC_CLIENT)
            cscope.span.set_tag(ext.PEER_HOSTNAME, "localhost")
            cscope.span.set_tag(ext.HTTP_URL, "/python/simple/two")
            cscope.span.set_tag(ext.HTTP_METHOD, "POST")
            cscope.span.set_tag(ext.HTTP_STATUS_CODE, 204)
            cscope.span.set_baggage_item("someBaggage", "someValue")
            time.sleep(.1)

    return HttpResponse('Stan wuz here!')


urlpatterns = [
    url(r'^$', index, name='index'),
    url(r'^cause_error$', cause_error, name='cause_error'),
    url(r'^another$', another),
    url(r'^not_found$', not_found, name='not_found'),
    url(r'^complex$', complex, name='complex')
]



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

from instana.log import logger
from instana.propagators.base_propagator import BasePropagator


class TextPropagator(BasePropagator):
    """
    Instana context propagator for TEXT_MAP.

    The TEXT_MAP deals with key-values with string to string mapping.
    The character set is unrestricted.
    """

    def inject(self, span_context, carrier, disable_w3c_trace_context=True):
        try:
            trace_id = span_context.trace_id
            span_id = span_context.span_id

            if isinstance(carrier, dict) or hasattr(carrier, "__dict__"):
                carrier[self.LC_HEADER_KEY_T] = trace_id
                carrier[self.LC_HEADER_KEY_S] = span_id
                carrier[self.LC_HEADER_KEY_L] = "1"
            elif isinstance(carrier, list):
                carrier.append((self.LC_HEADER_KEY_T, trace_id))
                carrier.append((self.LC_HEADER_KEY_S, span_id))
                carrier.append((self.LC_HEADER_KEY_L, "1"))
            elif isinstance(carrier, tuple):
                carrier = carrier.__add__(((self.LC_HEADER_KEY_T, trace_id),))
                carrier = carrier.__add__(((self.LC_HEADER_KEY_S, span_id),))
                carrier = carrier.__add__(((self.LC_HEADER_KEY_L, "1"),))
            elif hasattr(carrier, '__setitem__'):
                carrier.__setitem__(self.LC_HEADER_KEY_T, trace_id)
                carrier.__setitem__(self.LC_HEADER_KEY_S, span_id)
                carrier.__setitem__(self.LC_HEADER_KEY_L, "1")
            else:
                raise Exception("Unsupported carrier type", type(carrier))

            return carrier
        except Exception:
            logger.debug("inject error:", exc_info=True)
